# -*- coding: utf-8 -*-

"""Utilities for handling nodes."""

import itertools as itt
import logging
from itertools import chain
from typing import Set, Tuple, Type, Union

from networkx import relabel_nodes

from ..constants import ANNOTATIONS, CITATION, EVIDENCE, INCREASES, RELATION
from ..dsl import BaseAbundance, BaseEntity, ListAbundance, Reaction

__all__ = [
    "flatten_list_abundance",
    "list_abundance_cartesian_expansion",
    "reaction_cartesian_expansion",
]

logger = logging.getLogger(__name__)


def flatten_list_abundance(node: ListAbundance) -> ListAbundance:
    """Flattens the complex or composite abundance."""
    return node.__class__(
        list(
            chain.from_iterable(
                (flatten_list_abundance(member).members if isinstance(member, ListAbundance) else [member])
                for member in node.members
            )
        )
    )


def list_abundance_expansion(graph) -> None:
    """Flatten list abundances."""
    mapping = {node: flatten_list_abundance(node) for node in graph if isinstance(node, ListAbundance)}
    relabel_nodes(graph, mapping, copy=False)


def list_abundance_cartesian_expansion(graph) -> None:
    """Expand all list abundances to simple subject-predicate-object networks."""
    for u, v, d in list(graph.edges(data=True)):
        if CITATION not in d:
            continue

        if isinstance(u, ListAbundance) and isinstance(v, ListAbundance):
            for u_member, v_member in itt.product(u.members, v.members):
                graph.add_qualified_edge(
                    u_member,
                    v_member,
                    relation=d[RELATION],
                    citation=d.get(CITATION),
                    evidence=d.get(EVIDENCE),
                    annotations=d.get(ANNOTATIONS),
                )

        elif isinstance(u, ListAbundance):
            for member in u.members:
                graph.add_qualified_edge(
                    member,
                    v,
                    relation=d[RELATION],
                    citation=d.get(CITATION),
                    evidence=d.get(EVIDENCE),
                    annotations=d.get(ANNOTATIONS),
                )

        elif isinstance(v, ListAbundance):
            for member in v.members:
                graph.add_qualified_edge(
                    u,
                    member,
                    relation=d[RELATION],
                    citation=d.get(CITATION),
                    evidence=d.get(EVIDENCE),
                    annotations=d.get(ANNOTATIONS),
                )

    _remove_list_abundance_nodes(graph)


def _reaction_cartesian_expansion_unqualified_helper(
    graph,
    u: BaseEntity,
    v: BaseEntity,
    d: dict,
) -> None:
    """Help deal with cartesian expansion in unqualified edges."""
    if isinstance(u, Reaction) and isinstance(v, Reaction):
        enzymes = _get_catalysts_in_reaction(u) | _get_catalysts_in_reaction(v)

        for reactant, product in chain(itt.product(u.reactants, u.products), itt.product(v.reactants, v.products)):
            if reactant in enzymes or product in enzymes:
                continue

            graph.add_unqualified_edge(reactant, product, INCREASES)

        for product, reactant in itt.product(u.products, u.reactants):
            if reactant in enzymes or product in enzymes:
                continue

            graph.add_unqualified_edge(
                product,
                reactant,
                d[RELATION],
            )

    elif isinstance(u, Reaction):
        enzymes = _get_catalysts_in_reaction(u)

        for product in u.products:
            # Skip create increases edges between enzymes
            if product in enzymes:
                continue

            # Only add edge between v and reaction if the node is not part of the reaction
            # In practice skips hasReactant, hasProduct edges
            if v not in u.products and v not in u.reactants:
                graph.add_unqualified_edge(product, v, INCREASES)
            for reactant in u.reactants:
                graph.add_unqualified_edge(reactant, product, INCREASES)

    elif isinstance(v, Reaction):
        enzymes = _get_catalysts_in_reaction(v)

        for reactant in v.reactants:
            # Skip create increases edges between enzymes
            if reactant in enzymes:
                continue

            # Only add edge between v and reaction if the node is not part of the reaction
            # In practice skips hasReactant, hasProduct edges
            if u not in v.products and u not in v.reactants:
                graph.add_unqualified_edge(u, reactant, INCREASES)
            for product in v.products:
                graph.add_unqualified_edge(reactant, product, INCREASES)


def _get_catalysts_in_reaction(reaction: Reaction) -> Set[BaseAbundance]:
    """Return nodes that are both in reactants and reactions in a reaction."""
    # TODO replace with reaction.get_catalysts()
    return set(reaction.reactants).intersection(reaction.products)


def reaction_cartesian_expansion(graph, accept_unqualified_edges: bool = True) -> None:
    """Expand all reactions to simple subject-predicate-object networks."""
    for u, v, d in list(graph.edges(data=True)):
        # Deal with unqualified edges
        if CITATION not in d and accept_unqualified_edges:
            _reaction_cartesian_expansion_unqualified_helper(graph, u, v, d)
            continue

        if isinstance(u, Reaction) and isinstance(v, Reaction):
            catalysts = _get_catalysts_in_reaction(u) | _get_catalysts_in_reaction(v)

            for reactant, product in chain(
                itt.product(u.reactants, u.products),
                itt.product(v.reactants, v.products),
            ):
                if reactant in catalysts or product in catalysts:
                    continue
                graph.add_increases(
                    reactant,
                    product,
                    citation=d.get(CITATION),
                    evidence=d.get(EVIDENCE),
                    annotations=d.get(ANNOTATIONS),
                )

            for product, reactant in itt.product(u.products, u.reactants):
                if reactant in catalysts or product in catalysts:
                    continue

                graph.add_qualified_edge(
                    product,
                    reactant,
                    relation=d[RELATION],
                    citation=d.get(CITATION),
                    evidence=d.get(EVIDENCE),
                    annotations=d.get(ANNOTATIONS),
                )

        elif isinstance(u, Reaction):
            catalysts = _get_catalysts_in_reaction(u)

            for product in u.products:
                # Skip create increases edges between enzymes
                if product in catalysts:
                    continue

                # Only add edge between v and reaction if the node is not part of the reaction
                # In practice skips hasReactant, hasProduct edges
                if v not in u.products and v not in u.reactants:
                    graph.add_increases(
                        product,
                        v,
                        citation=d.get(CITATION),
                        evidence=d.get(EVIDENCE),
                        annotations=d.get(ANNOTATIONS),
                    )

                for reactant in u.reactants:
                    graph.add_increases(
                        reactant,
                        product,
                        citation=d.get(CITATION),
                        evidence=d.get(EVIDENCE),
                        annotations=d.get(ANNOTATIONS),
                    )

        elif isinstance(v, Reaction):
            catalysts = _get_catalysts_in_reaction(v)

            for reactant in v.reactants:
                # Skip create increases edges between enzymes
                if reactant in catalysts:
                    continue

                # Only add edge between v and reaction if the node is not part of the reaction
                # In practice skips hasReactant, hasProduct edges
                if u not in v.products and u not in v.reactants:
                    graph.add_increases(
                        u,
                        reactant,
                        citation=d.get(CITATION),
                        evidence=d.get(EVIDENCE),
                        annotations=d.get(ANNOTATIONS),
                    )
                for product in v.products:
                    graph.add_increases(
                        reactant,
                        product,
                        citation=d.get(CITATION),
                        evidence=d.get(EVIDENCE),
                        annotations=d.get(ANNOTATIONS),
                    )

    _remove_reaction_nodes(graph)


def remove_reified_nodes(graph) -> None:
    """Remove complex nodes."""
    _remove_list_abundance_nodes(graph)
    _remove_reaction_nodes(graph)


def _remove_list_abundance_nodes(graph):
    _remove_typed_nodes(graph, ListAbundance)


def _remove_reaction_nodes(graph):
    _remove_typed_nodes(graph, Reaction)


def _remove_typed_nodes(
    graph,
    cls: Union[Type[BaseEntity], Tuple[Type[BaseEntity], ...]],
) -> None:
    graph.remove_nodes_from({node for node in graph if isinstance(node, cls)})



# -*- coding: utf-8 -*-

"""Managers and subclasses for PyBEL.

The :mod:`pybel.manager` module serves as an interface between the BEL graph data structure and underlying relational
databases. Its inclusion allows for the caching of namespaces and annotations for much faster lookup than
downloading and parsing upon each compilation.
"""

from . import (
    base_manager,
    cache_manager,
    citation_utils,
    database_io,
    make_json_serializable,
    models,
    query_manager,
)
from .base_manager import BaseManager, build_engine_session
from .cache_manager import Manager, NetworkManager
from .citation_utils import enrich_pmc_citations, enrich_pubmed_citations
from .database_io import from_database, to_database
from .models import (
    Author,
    Base,
    Citation,
    Edge,
    Evidence,
    Namespace,
    NamespaceEntry,
    Network,
    Node,
    edge_annotation,
    network_edge,
    network_node,
)
from .query_manager import QueryManager, graph_from_edges

__all__ = [
    "BaseManager",
    "build_engine_session",
    "Manager",
    "NetworkManager",
    "QueryManager",
    "graph_from_edges",
    "enrich_pubmed_citations",
    "enrich_pmc_citations",
    # I/O
    "from_database",
    "to_database",
    # Models
    "Base",
    "Namespace",
    "NamespaceEntry",
    "Network",
    "Node",
    "Author",
    "Citation",
    "Evidence",
    "Edge",
    "edge_annotation",
    "network_edge",
    "network_node",
]



# -*- coding: utf-8 -*-

"""Tests for parsing full BEL relations."""

import logging
import unittest

from pyparsing import ParseException

from pybel import BELGraph
from pybel.canonicalize import edge_to_bel
from pybel.constants import (
    ABUNDANCE,
    ACTIVITY,
    ANNOTATIONS,
    BIOPROCESS,
    CAUSES_NO_CHANGE,
    CITATION,
    COMPLEX,
    COMPOSITE,
    CONCEPT,
    CORRELATION,
    DECREASES,
    DIRECTLY_DECREASES,
    DIRECTLY_INCREASES,
    EFFECT,
    EQUIVALENT_TO,
    EVIDENCE,
    FROM_LOC,
    FUNCTION,
    GENE,
    GMOD,
    GRAPH_ANNOTATION_LIST,
    HAS_PRODUCT,
    HAS_REACTANT,
    HAS_VARIANT,
    HGVS,
    IDENTIFIER,
    INCREASES,
    IS_A,
    KIND,
    LOCATION,
    MEMBERS,
    MODIFIER,
    NAME,
    NAMESPACE,
    NEGATIVE_CORRELATION,
    NO_CORRELATION,
    ORTHOLOGOUS,
    PART_OF,
    PATHOLOGY,
    POSITIVE_CORRELATION,
    PRODUCTS,
    PROTEIN,
    RATE_LIMITING_STEP_OF,
    REACTANTS,
    REACTION,
    REGULATES,
    RELATION,
    RNA,
    SOURCE,
    SOURCE_MODIFIER,
    SUBPROCESS_OF,
    TARGET,
    TARGET_MODIFIER,
    TO_LOC,
    TRANSCRIBED_TO,
    TRANSLATED_TO,
    TRANSLOCATION,
    VARIANTS,
)
from pybel.dsl import (
    ComplexAbundance,
    Pathology,
    Protein,
    Rna,
    abundance,
    activity,
    bioprocess,
    complex_abundance,
    composite_abundance,
    gene,
    gmod,
    hgvs,
    named_complex_abundance,
    pmod,
    protein,
    reaction,
    rna,
)
from pybel.dsl.namespaces import hgnc
from pybel.exceptions import (
    MissingNamespaceNameWarning,
    NestedRelationWarning,
    UndefinedNamespaceWarning,
)
from pybel.language import Entity, activity_mapping
from pybel.parser import BELParser
from tests.constants import TestTokenParserBase, test_citation_dict, test_evidence_text

logger = logging.getLogger(__name__)


class TestRelations(TestTokenParserBase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.parser.relation.streamline()

    def setUp(self):
        super().setUp()
        self.add_default_provenance()

    def test_ensure_no_dup_nodes(self):
        """Ensure node isn't added twice, even if from different statements"""
        self.parser.gene.addParseAction(self.parser.handle_term)
        result = self.parser.bel_term.parseString("g(HGNC:AKT1)")

        expected_result_dict = {
            FUNCTION: GENE,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
        }

        self.assertEqual(expected_result_dict, result.asDict())

        self.parser.degradation.addParseAction(self.parser.handle_term)
        self.parser.degradation.parseString("deg(g(HGNC:AKT1))")

        akt1_gene = gene("HGNC", "AKT1")

        self.assertEqual(1, self.parser.graph.number_of_nodes())
        self.assert_has_node(akt1_gene)

    def test_singleton(self):
        """Test singleton composite in subject."""
        statement = 'composite(p(HGNC:CASP8),p(HGNC:FADD),a(CHEBI:"Abeta_42"))'
        result = self.parser.statement.parseString(statement)

        expected = [
            COMPOSITE,
            [PROTEIN, ["HGNC", "CASP8"]],
            [PROTEIN, ["HGNC", "FADD"]],
            [ABUNDANCE, ["CHEBI", "Abeta_42"]],
        ]
        self.assertEqual(expected, result.asList())

        sub_member_1 = protein("HGNC", "CASP8")
        self.assert_has_node(sub_member_1)

        sub_member_2 = protein("HGNC", "FADD")
        self.assert_has_node(sub_member_2)

        sub_member_3 = abundance("CHEBI", "Abeta_42")
        self.assert_has_node(sub_member_3)

        sub = composite_abundance([sub_member_1, sub_member_2, sub_member_3])
        self.assert_has_node(sub)

        self.assert_has_edge(sub_member_1, sub, relation=PART_OF)
        self.assert_has_edge(sub_member_2, sub, relation=PART_OF)

    def test_predicate_failure(self):
        """Checks that if there's a problem with the relation/object, that an error gets thrown"""
        statement = 'composite(p(HGNC:CASP8),p(HGNC:FADD),a(CHEBI:"Abeta_42")) -> nope(GO:"neuron apoptotic process")'

        with self.assertRaises(ParseException):
            self.parser.relation.parseString(statement)

    def test_increases(self):
        """Test composite in subject. See BEL 2.0 specification
        `3.1.1 <http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#Xincreases>`_
        """
        statement = 'composite(p(HGNC:CASP8),p(HGNC:FADD),a(CHEBI:"Abeta_42")) -> bp(GO:"neuron apoptotic process")'
        result = self.parser.relation.parseString(statement)

        expected = [
            [
                COMPOSITE,
                [PROTEIN, ["HGNC", "CASP8"]],
                [PROTEIN, ["HGNC", "FADD"]],
                [ABUNDANCE, ["CHEBI", "Abeta_42"]],
            ],
            INCREASES,
            [BIOPROCESS, ["GO", "neuron apoptotic process"]],
        ]
        self.assertEqual(expected, result.asList())

        sub_member_1 = protein("HGNC", "CASP8")
        self.assert_has_node(sub_member_1)

        sub_member_2 = protein("HGNC", "FADD")
        self.assert_has_node(sub_member_2)

        sub_member_3 = abundance("CHEBI", "Abeta_42")
        self.assert_has_node(sub_member_3)

        sub = composite_abundance([sub_member_1, sub_member_2, sub_member_3])
        self.assert_has_node(sub)

        self.assert_has_edge(sub_member_1, sub, relation=PART_OF)
        self.assert_has_edge(sub_member_2, sub, relation=PART_OF)
        self.assert_has_edge(sub_member_3, sub, relation=PART_OF)

        obj = bioprocess("GO", "neuron apoptotic process")
        self.assert_has_node(obj)

        self.assert_has_edge(sub, obj, relation=INCREASES)

    def test_increases_methylation(self):
        """Test a causal statement with a gene modification."""
        for gmod in ["Me", 'go:0006306 ! "DNA methylation"']:
            with self.subTest(gmod=gmod):
                self._help_test_increases_methylation(gmod)

    def _help_test_increases_methylation(self, x):
        statement = f'a(CHEBI:"lead atom") -> g(HGNC:APP, gmod({x}))'
        result = self.parser.relation.parseString(statement)
        expected_dict = {
            TARGET: {
                FUNCTION: GENE,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "APP",
                },
                VARIANTS: [
                    {
                        KIND: GMOD,
                        CONCEPT: {
                            NAMESPACE: "go",
                            IDENTIFIER: "0006306",
                            NAME: "DNA methylation",
                        },
                    },
                ],
            },
            RELATION: INCREASES,
            SOURCE: {
                FUNCTION: ABUNDANCE,
                CONCEPT: {
                    NAMESPACE: "CHEBI",
                    NAME: "lead atom",
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        sub = abundance("CHEBI", "lead atom")
        obj = gene("HGNC", "APP", variants=gmod("Me"))

        self.assert_has_edge(sub, obj, relation=INCREASES)

    def test_directlyIncreases_withTlocObject(self):
        """Test translocation in object. See BEL 2.0 specification
        `3.1.2 <http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#XdIncreases>`_
        """
        statement = (
            'a(CHEBI:"Abeta_42") => tloc(a(CHEBI:"calcium(2+)"),fromLoc(MESH:"Cell Membrane"),'
            'toLoc(MESH:"Intracellular Space"))'
        )
        result = self.parser.relation.parseString(statement)

        expected_dict = {
            SOURCE: {
                FUNCTION: ABUNDANCE,
                CONCEPT: {
                    NAMESPACE: "CHEBI",
                    NAME: "Abeta_42",
                },
            },
            RELATION: DIRECTLY_INCREASES,
            TARGET: {
                FUNCTION: ABUNDANCE,
                CONCEPT: {
                    NAMESPACE: "CHEBI",
                    NAME: "calcium(2+)",
                },
                MODIFIER: TRANSLOCATION,
                EFFECT: {
                    FROM_LOC: {NAMESPACE: "MESH", NAME: "Cell Membrane"},
                    TO_LOC: {NAMESPACE: "MESH", NAME: "Intracellular Space"},
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        sub = abundance("CHEBI", "Abeta_42")
        self.assert_has_node(sub)

        obj = abundance("CHEBI", "calcium(2+)")
        self.assert_has_node(obj)

        expected_annotations = {
            RELATION: DIRECTLY_INCREASES,
            TARGET_MODIFIER: {
                MODIFIER: TRANSLOCATION,
                EFFECT: {
                    FROM_LOC: {NAMESPACE: "MESH", NAME: "Cell Membrane"},
                    TO_LOC: {NAMESPACE: "MESH", NAME: "Intracellular Space"},
                },
            },
        }

        self.assert_has_edge(sub, obj, **expected_annotations)

    def test_decreases(self):
        """Test parsing a decreases relation with a reaction.

        3.1.3 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#Xdecreases
        """
        statement = "pep(p(FPLX:CAPN, location(GO:intracellular))) -| reaction(reactants(p(HGNC:CDK5R1)),products(p(HGNC:CDK5)))"
        result = self.parser.relation.parseString(statement)

        expected_dict = {
            SOURCE: {
                MODIFIER: ACTIVITY,
                FUNCTION: PROTEIN,
                CONCEPT: {
                    NAMESPACE: "FPLX",
                    NAME: "CAPN",
                },
                LOCATION: {NAMESPACE: "GO", NAME: "intracellular"},
                EFFECT: activity_mapping["pep"],
            },
            RELATION: DECREASES,
            TARGET: {
                FUNCTION: REACTION,
                REACTANTS: [
                    {
                        FUNCTION: PROTEIN,
                        CONCEPT: {
                            NAMESPACE: "HGNC",
                            NAME: "CDK5R1",
                        },
                    }
                ],
                PRODUCTS: [
                    {
                        FUNCTION: PROTEIN,
                        CONCEPT: {
                            NAMESPACE: "HGNC",
                            NAME: "CDK5",
                        },
                    },
                ],
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        sub = protein("FPLX", "CAPN")
        self.assert_has_node(sub)

        obj_member_1 = protein("HGNC", "CDK5R1")
        self.assert_has_node(obj_member_1)

        obj_member_2 = protein("HGNC", "CDK5")
        self.assert_has_node(obj_member_2)

        obj = reaction(reactants=[obj_member_1], products=[obj_member_2])
        self.assert_has_node(obj)

        self.assert_has_edge(obj, obj_member_1, relation=HAS_REACTANT)
        self.assert_has_edge(obj, obj_member_2, relation=HAS_PRODUCT)

        expected_edge_attributes = {
            RELATION: DECREASES,
            SOURCE_MODIFIER: {
                MODIFIER: ACTIVITY,
                EFFECT: activity_mapping["pep"],
                LOCATION: {
                    NAMESPACE: "GO",
                    NAME: "intracellular",
                },
            },
        }

        self.assertEqual(
            expected_edge_attributes[SOURCE_MODIFIER],
            activity(name="pep", location=Entity(name="intracellular", namespace="GO")),
        )

        self.assert_has_edge(sub, obj, **expected_edge_attributes)

    def test_directlyDecreases(self):
        """
        3.1.4 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#XdDecreases
        Tests simple triple"""
        statement = 'proteinAbundance(HGNC:CAT, location(GO:intracellular)) directlyDecreases abundance(CHEBI:"hydrogen peroxide")'
        result = self.parser.relation.parseString(statement)

        expected_dict = {
            SOURCE: {
                FUNCTION: PROTEIN,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "CAT",
                },
                LOCATION: {NAMESPACE: "GO", NAME: "intracellular"},
            },
            RELATION: DIRECTLY_DECREASES,
            TARGET: {
                FUNCTION: ABUNDANCE,
                CONCEPT: {
                    NAMESPACE: "CHEBI",
                    NAME: "hydrogen peroxide",
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        sub = protein("HGNC", "CAT")
        self.assert_has_node(sub)

        obj = abundance("CHEBI", "hydrogen peroxide")
        self.assert_has_node(obj)

        expected_attrs = {
            SOURCE_MODIFIER: {LOCATION: {NAMESPACE: "GO", NAME: "intracellular"}},
            RELATION: DIRECTLY_DECREASES,
        }
        self.assert_has_edge(sub, obj, **expected_attrs)

    def test_directlyDecreases_annotationExpansion(self):
        """
        3.1.4 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#XdDecreases
        Tests simple triple"""
        statement = 'g(HGNC:CAT, location(GO:intracellular)) directlyDecreases abundance(CHEBI:"hydrogen peroxide")'

        self.graph.annotation_list.update(
            {
                "ListAnnotation": set("abef"),
                "ScalarAnnotation": set("cghi"),
            }
        )
        annotations = self.parser.graph._clean_annotations(
            {
                "ListAnnotation": {"a", "b"},
                "ScalarAnnotation": {"c"},
            }
        )

        self.parser.control_parser.annotations.update(annotations)

        result = self.parser.relation.parseString(statement)

        expected_dict = {
            SOURCE: {
                FUNCTION: GENE,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "CAT",
                },
                LOCATION: {
                    NAMESPACE: "GO",
                    NAME: "intracellular",
                },
            },
            RELATION: DIRECTLY_DECREASES,
            TARGET: {
                FUNCTION: ABUNDANCE,
                CONCEPT: {
                    NAMESPACE: "CHEBI",
                    NAME: "hydrogen peroxide",
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        sub = gene("HGNC", "CAT")
        self.assert_has_node(sub)

        obj = abundance("CHEBI", "hydrogen peroxide")
        self.assert_has_node(obj)

        expected_attrs = {
            SOURCE_MODIFIER: {LOCATION: {NAMESPACE: "GO", NAME: "intracellular"}},
            RELATION: DIRECTLY_DECREASES,
            CITATION: test_citation_dict,
            EVIDENCE: test_evidence_text,
            ANNOTATIONS: {
                "ListAnnotation": [
                    Entity(namespace="ListAnnotation", identifier="a"),
                    Entity(namespace="ListAnnotation", identifier="b"),
                ],
                "ScalarAnnotation": [
                    Entity(namespace="ScalarAnnotation", identifier="c"),
                ],
            },
        }
        self.assert_has_edge(sub, obj, only=True, **expected_attrs)

    def test_rateLimitingStepOf_subjectActivity(self):
        """3.1.5 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#_ratelimitingstepof"""
        statement = 'act(p(HGNC:HMGCR), ma(cat)) rateLimitingStepOf bp(GO:"cholesterol biosynthetic process")'
        result = self.parser.relation.parseString(statement)

        expected_dict = {
            SOURCE: {
                MODIFIER: ACTIVITY,
                FUNCTION: PROTEIN,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "HMGCR",
                },
                EFFECT: activity_mapping["cat"],
            },
            RELATION: RATE_LIMITING_STEP_OF,
            TARGET: {
                FUNCTION: BIOPROCESS,
                CONCEPT: {
                    NAMESPACE: "GO",
                    NAME: "cholesterol biosynthetic process",
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        sub = protein("HGNC", "HMGCR")
        self.assert_has_node(sub)

        obj = bioprocess("GO", "cholesterol biosynthetic process")
        self.assert_has_node(obj)

        self.assert_has_edge(sub, obj, relation=expected_dict[RELATION])

    def test_cnc_with_subject_variant(self):
        """Test a causesNoChange relationship with a variant in the subject.

        See also: 3.1.6 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#Xcnc
        """
        statement = 'g(HGNC:APP,sub(G,275341,C)) cnc path(MESH:"Alzheimer Disease")'
        result = self.parser.relation.parseString(statement)

        expected_dict = {
            SOURCE: {
                FUNCTION: GENE,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "APP",
                },
                VARIANTS: [
                    {KIND: HGVS, HGVS: "c.275341G>C"},
                ],
            },
            RELATION: CAUSES_NO_CHANGE,
            TARGET: {
                FUNCTION: PATHOLOGY,
                CONCEPT: {
                    NAMESPACE: "MESH",
                    NAME: "Alzheimer Disease",
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        app_gene = gene(namespace="HGNC", name="APP")
        self.assert_has_node(app_gene)
        sub = app_gene.with_variants(hgvs("c.275341G>C"))
        self.assert_has_node(sub)

        obj = Pathology("MESH", "Alzheimer Disease")
        self.assert_has_node(obj)

        self.assert_has_edge(sub, obj, relation=expected_dict[RELATION])

    def test_regulates_with_multiple_annotations(self):
        """
        3.1.7 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#_regulates_reg
        Test nested definitions"""
        statement = "pep(complex(p(HGNC:F3),p(HGNC:F7))) regulates pep(p(HGNC:F9))"
        result = self.parser.relation.parseString(statement)

        expected_dict = {
            SOURCE: {
                MODIFIER: ACTIVITY,
                EFFECT: activity_mapping["pep"],
                FUNCTION: COMPLEX,
                MEMBERS: [
                    {FUNCTION: PROTEIN, CONCEPT: {NAMESPACE: "HGNC", NAME: "F3"}},
                    {FUNCTION: PROTEIN, CONCEPT: {NAMESPACE: "HGNC", NAME: "F7"}},
                ],
            },
            RELATION: REGULATES,
            TARGET: {
                MODIFIER: ACTIVITY,
                EFFECT: activity_mapping["pep"],
                FUNCTION: PROTEIN,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "F9",
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        sub_member_1 = protein("HGNC", "F3")
        self.assert_has_node(sub_member_1)

        sub_member_2 = protein("HGNC", "F7")
        self.assert_has_node(sub_member_2)

        sub = complex_abundance([sub_member_1, sub_member_2])
        self.assert_has_node(sub)

        self.assert_has_edge(sub_member_1, sub, relation=PART_OF)
        self.assert_has_edge(sub_member_2, sub, relation=PART_OF)

        obj = protein("HGNC", "F9")
        self.assert_has_node(obj)

        self.assert_has_edge(sub, obj, relation=expected_dict[RELATION])

    def test_nested_failure(self):
        """Test nested statement (3.1)."""
        statement = 'p(HGNC:CAT) -| (a(CHEBI:"hydrogen peroxide") -> bp(GO:"apoptotic process"))'
        self.parser.disallow_nested = True
        with self.assertRaises(NestedRelationWarning):
            self.parser.relation.parseString(statement)
        self.parser.disallow_nested = False

    def test_nested_lenient(self):
        """Test nested statement (3.1)."""
        self.parser.disallow_nested = False
        statement = 'p(HGNC:CAT) -| (a(CHEBI:"hydrogen peroxide") -> bp(GO:"apoptotic process"))'
        self.parser.relation.parseString(statement)

        cat = protein("HGNC", "CAT")
        h2o2 = abundance("CHEBI", "hydrogen peroxide")
        apoptosis = bioprocess("GO", "apoptotic process")

        self.assert_has_edge(cat, h2o2)
        self.assert_has_edge(h2o2, apoptosis)
        self.assertEqual(1, len(self.parser.graph.transitivities))

    def test_negativeCorrelation_withObjectVariant(self):
        """
        3.2.1 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#XnegCor
        Test phosphoralation tag"""
        for pmod in ["P", "Ph", 'go:0006468 ! "protein phosphorylation"']:
            with self.subTest(pmod=pmod):
                self._help_test_negative_correlation_with_object_variant(pmod)

    def _help_test_negative_correlation_with_object_variant(self, x):
        statement = f"kin(p(FPLX:GSK3)) neg p(HGNC:MAPT,pmod({x}))"
        result = self.parser.relation.parseString(statement)

        expected_dict = {
            SOURCE: {
                MODIFIER: ACTIVITY,
                EFFECT: activity_mapping["kin"],
                FUNCTION: PROTEIN,
                CONCEPT: {
                    NAMESPACE: "FPLX",
                    NAME: "GSK3",
                },
            },
            RELATION: NEGATIVE_CORRELATION,
            TARGET: {
                FUNCTION: PROTEIN,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "MAPT",
                },
                VARIANTS: [pmod("Ph")],
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        sub = protein("FPLX", "GSK3")
        self.assert_has_node(sub)

        obj = protein("HGNC", "MAPT", variants=pmod("Ph"))
        self.assert_has_node(obj)

        self.assert_has_edge(sub, obj, relation=expected_dict[RELATION])
        self.assert_has_edge(obj, sub, relation=expected_dict[RELATION])

    def test_positiveCorrelation_withSelfReferential(self):
        """
        3.2.2 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#XposCor
        Self-referential relationships"""
        statement = "p(HGNC:GSK3B, pmod(P, S, 9)) pos act(p(HGNC:GSK3B), ma(kin))"
        result = self.parser.relation.parseString(statement)

        expected_dict = {
            SOURCE: {
                FUNCTION: PROTEIN,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "GSK3B",
                },
                VARIANTS: [pmod("Ph", position=9, code="Ser")],
            },
            RELATION: POSITIVE_CORRELATION,
            TARGET: {
                MODIFIER: ACTIVITY,
                FUNCTION: PROTEIN,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "GSK3B",
                },
                EFFECT: activity_mapping["kin"],
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        subject_node = protein("HGNC", "GSK3B", variants=pmod("Ph", code="Ser", position=9))
        self.assert_has_node(subject_node)

        object_node = protein("HGNC", "GSK3B")
        self.assert_has_node(object_node)

        self.assert_has_edge(subject_node, object_node, relation=expected_dict[RELATION])
        self.assert_has_edge(object_node, subject_node, relation=expected_dict[RELATION])

    def test_orthologous(self):
        """
        3.3.1 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#_orthologous
        """
        statement = "g(HGNC:AKT1) orthologous g(MGI:AKT1)"
        result = self.parser.relation.parseString(statement)
        expected_result = [
            [GENE, ["HGNC", "AKT1"]],
            ORTHOLOGOUS,
            [GENE, ["MGI", "AKT1"]],
        ]
        self.assertEqual(expected_result, result.asList())

        sub = gene("HGNC", "AKT1")
        self.assert_has_node(sub)

        obj = gene("MGI", "AKT1")
        self.assert_has_node(obj)

        self.assert_has_edge(sub, obj, relation=ORTHOLOGOUS)
        self.assert_has_edge(obj, sub, relation=ORTHOLOGOUS)

    def test_transcription(self):
        """
        3.3.2 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#_transcribedto
        """
        statement = "g(HGNC:AKT1) :> r(HGNC:AKT1)"
        result = self.parser.relation.parseString(statement)

        expected_result = [
            [GENE, ["HGNC", "AKT1"]],
            TRANSCRIBED_TO,
            [RNA, ["HGNC", "AKT1"]],
        ]
        self.assertEqual(expected_result, result.asList())

        sub = gene("HGNC", "AKT1")
        self.assert_has_node(sub)

        obj = rna("HGNC", "AKT1")
        self.assert_has_node(obj)

        self.assert_has_edge(sub, obj, **{RELATION: TRANSCRIBED_TO})

    def test_translation(self):
        """
        3.3.3 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#_translatedto
        """
        statement = "r(HGNC:AKT1,loc(GO:intracellular)) >> p(HGNC:AKT1)"
        result = self.parser.relation.parseString(statement)

        # [[RNA, ['HGNC', 'AKT1']], TRANSLATED_TO, [PROTEIN, ['HGNC', 'AKT1']]]
        expected_result = {
            SOURCE: {
                FUNCTION: RNA,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "AKT1",
                },
                LOCATION: {
                    NAMESPACE: "GO",
                    NAME: "intracellular",
                },
            },
            RELATION: TRANSLATED_TO,
            TARGET: {
                FUNCTION: PROTEIN,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "AKT1",
                },
            },
        }
        self.assertEqual(expected_result, result.asDict())

        self.assertEqual(2, self.graph.number_of_nodes())

        source = rna(name="AKT1", namespace="HGNC")
        self.assertIn(source, self.graph)

        target = protein(name="AKT1", namespace="HGNC")
        self.assertIn(target, self.graph)

        self.assertEqual(1, self.graph.number_of_edges())
        self.assertTrue(self.graph.has_edge(source, target))

        key_data = self.parser.graph[source][target]
        self.assertEqual(1, len(key_data))

        key = list(key_data)[0]
        data = key_data[key]

        self.assertIn(RELATION, data)
        self.assertEqual(TRANSLATED_TO, data[RELATION])

        calculated_edge_bel = edge_to_bel(source, target, data=data)
        self.assertEqual(
            "r(HGNC:AKT1, loc(GO:intracellular)) translatedTo p(HGNC:AKT1)",
            calculated_edge_bel,
        )

    def test_component_list(self):
        s = "complex(FPLX:C1) hasComponents list(p(HGNC:C1QB), p(HGNC:C1S))"
        result = self.parser.relation.parseString(s)

        expected_result_list = [
            [COMPLEX, ["FPLX", "C1"]],
            "hasComponents",
            [
                [PROTEIN, ["HGNC", "C1QB"]],
                [PROTEIN, ["HGNC", "C1S"]],
            ],
        ]
        self.assertEqual(expected_result_list, result.asList())

        sub = named_complex_abundance("FPLX", "C1")
        self.assert_has_node(sub)
        child_1 = hgnc(name="C1QB")
        self.assert_has_node(child_1)
        self.assert_has_edge(child_1, sub, **{RELATION: PART_OF})
        child_2 = hgnc(name="C1S")
        self.assert_has_node(child_2)
        self.assert_has_edge(child_2, sub, **{RELATION: PART_OF})

    def test_member_list(self):
        """
        3.4.2 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#_hasmembers
        """
        statement = "p(FPLX:PKC) hasMembers list(p(HGNC:PRKCA), p(HGNC:PRKCB), p(HGNC:PRKCD), p(HGNC:PRKCE))"
        result = self.parser.relation.parseString(statement)
        expected_result = [
            [PROTEIN, ["FPLX", "PKC"]],
            "hasMembers",
            [
                [PROTEIN, ["HGNC", "PRKCA"]],
                [PROTEIN, ["HGNC", "PRKCB"]],
                [PROTEIN, ["HGNC", "PRKCD"]],
                [PROTEIN, ["HGNC", "PRKCE"]],
            ],
        ]
        self.assertEqual(expected_result, result.asList())

        sub = protein("FPLX", "PKC")
        obj_1 = protein("HGNC", "PRKCA")
        obj_2 = protein("HGNC", "PRKCB")
        obj_3 = protein("HGNC", "PRKCD")
        obj_4 = protein("HGNC", "PRKCE")

        self.assert_has_node(sub)

        self.assert_has_node(obj_1)
        self.assert_has_edge(obj_1, sub, relation=IS_A)

        self.assert_has_node(obj_2)
        self.assert_has_edge(obj_2, sub, relation=IS_A)

        self.assert_has_node(obj_3)
        self.assert_has_edge(obj_3, sub, relation=IS_A)

        self.assert_has_node(obj_4)
        self.assert_has_edge(obj_4, sub, relation=IS_A)

    def test_is_a(self):
        """
        3.4.5 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#_isa
        """
        statement = 'pathology(MESH:Psoriasis) isA pathology(MESH:"Skin Diseases")'
        result = self.parser.relation.parseString(statement)

        expected_result = [
            [PATHOLOGY, ["MESH", "Psoriasis"]],
            "isA",
            [PATHOLOGY, ["MESH", "Skin Diseases"]],
        ]
        self.assertEqual(expected_result, result.asList())

        sub = Pathology("MESH", "Psoriasis")
        self.assert_has_node(sub)

        obj = Pathology("MESH", "Skin Diseases")
        self.assert_has_node(obj)

        self.assert_has_edge(sub, obj, relation=IS_A)

    def test_equivalentTo(self):
        statement = 'g(dbSNP:"rs123456") eq g(HGNC:YFG, var(c.123G>A))'
        result = self.parser.relation.parseString(statement)

        expected_result = {
            SOURCE: {
                FUNCTION: GENE,
                CONCEPT: {
                    NAMESPACE: "dbSNP",
                    NAME: "rs123456",
                },
            },
            RELATION: EQUIVALENT_TO,
            TARGET: {
                FUNCTION: GENE,
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "YFG",
                },
                VARIANTS: [
                    {
                        KIND: HGVS,
                        HGVS: "c.123G>A",
                    },
                ],
            },
        }
        self.assertEqual(expected_result, result.asDict())

        sub = gene("dbSNP", "rs123456")
        self.assert_has_node(sub)

        obj = gene("HGNC", "YFG", variants=hgvs("c.123G>A"))
        self.assert_has_node(obj)

        self.assert_has_edge(sub, obj, **{RELATION: EQUIVALENT_TO})
        self.assert_has_edge(obj, sub, **{RELATION: EQUIVALENT_TO})

    def test_partOf(self):
        statement = 'a(UBERON:"corpus striatum") partOf a(UBERON:"basal ganglion")'
        self.parser.relation.parseString(statement)

        corpus_striatum = abundance(namespace="UBERON", name="corpus striatum")
        basal_ganglion = abundance(namespace="UBERON", name="basal ganglion")

        self.assert_has_node(corpus_striatum)
        self.assert_has_node(basal_ganglion)
        self.assert_has_edge(corpus_striatum, basal_ganglion, relation=PART_OF)

        v = list(self.parser.graph[corpus_striatum][basal_ganglion].values())
        self.assertEqual(1, len(v))

        v = v[0]
        self.assertIn(RELATION, v)
        self.assertEqual(PART_OF, v[RELATION])

    def test_subProcessOf(self):
        """
        3.4.6 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#_subprocessof
        """
        statement = 'rxn(reactants(a(CHEBI:"(S)-3-hydroxy-3-methylglutaryl-CoA"),a(CHEBI:NADPH), \
            a(CHEBI:hydron)),products(a(CHEBI:mevalonate), a(CHEBI:"CoA-SH"), a(CHEBI:"NADP(+)"))) \
            subProcessOf bp(GO:"cholesterol biosynthetic process")'
        result = self.parser.relation.parseString(statement)
        expected_result = [
            [
                REACTION,
                [
                    [ABUNDANCE, ["CHEBI", "(S)-3-hydroxy-3-methylglutaryl-CoA"]],
                    [ABUNDANCE, ["CHEBI", "NADPH"]],
                    [ABUNDANCE, ["CHEBI", "hydron"]],
                ],
                [
                    [ABUNDANCE, ["CHEBI", "mevalonate"]],
                    [ABUNDANCE, ["CHEBI", "CoA-SH"]],
                    [ABUNDANCE, ["CHEBI", "NADP(+)"]],
                ],
            ],
            SUBPROCESS_OF,
            [BIOPROCESS, ["GO", "cholesterol biosynthetic process"]],
        ]
        self.assertEqual(expected_result, result.asList())

        sub_reactant_1 = abundance("CHEBI", "(S)-3-hydroxy-3-methylglutaryl-CoA")
        sub_reactant_2 = abundance("CHEBI", "NADPH")
        sub_reactant_3 = abundance("CHEBI", "hydron")
        sub_product_1 = abundance("CHEBI", "mevalonate")
        sub_product_2 = abundance("CHEBI", "CoA-SH")
        sub_product_3 = abundance("CHEBI", "NADP(+)")

        self.assert_has_node(sub_reactant_1)
        self.assert_has_node(sub_reactant_2)
        self.assert_has_node(sub_reactant_3)
        self.assert_has_node(sub_product_1)
        self.assert_has_node(sub_product_2)
        self.assert_has_node(sub_product_3)

        sub = reaction(
            [sub_reactant_1, sub_reactant_2, sub_reactant_3],
            [sub_product_1, sub_product_2, sub_product_3],
        )

        self.assert_has_edge(sub, sub_reactant_1, relation=HAS_REACTANT)
        self.assert_has_edge(sub, sub_reactant_2, relation=HAS_REACTANT)
        self.assert_has_edge(sub, sub_reactant_3, relation=HAS_REACTANT)
        self.assert_has_edge(sub, sub_product_1, relation=HAS_PRODUCT)
        self.assert_has_edge(sub, sub_product_2, relation=HAS_PRODUCT)
        self.assert_has_edge(sub, sub_product_3, relation=HAS_PRODUCT)

        obj = bioprocess("GO", "cholesterol biosynthetic process")
        self.assert_has_node(obj)

        self.assert_has_edge(sub, obj, **{RELATION: SUBPROCESS_OF})

    def test_extra_1(self):
        statement = 'abundance(CHEBI:"nitric oxide") increases cellSurfaceExpression(complexAbundance(proteinAbundance(HGNC:ITGAV),proteinAbundance(HGNC:ITGB3)))'
        self.parser.relation.parseString(statement)

    def test_has_variant(self):
        statement = "g(HGNC:AKT1) hasVariant g(HGNC:AKT1, gmod(M))"
        self.parser.relation.parseString(statement)

        expected_parent = gene("HGNC", "AKT1")
        expected_child = expected_parent.with_variants(gmod("Me"))

        self.assert_has_node(expected_parent)
        self.assert_has_node(expected_child)

        self.assertEqual("g(HGNC:AKT1)", self.graph.node_to_bel(expected_parent))
        self.assertEqual(
            'g(HGNC:AKT1, gmod(go:0006306 ! "DNA methylation"))',
            self.graph.node_to_bel(expected_child),
        )

        self.assert_has_edge(expected_parent, expected_child, **{RELATION: HAS_VARIANT})

    def test_has_reaction_component(self):
        statement = 'rxn(reactants(a(CHEBI:"(S)-3-hydroxy-3-methylglutaryl-CoA"),a(CHEBI:NADPH), \
                    a(CHEBI:hydron)),products(a(CHEBI:mevalonate), a(CHEBI:"CoA-SH"), a(CHEBI:"NADP(+)"))) \
                    hasReactant a(CHEBI:"(S)-3-hydroxy-3-methylglutaryl-CoA")'
        self.parser.relation.parseString(statement)

        sub_reactant_1 = abundance("CHEBI", "(S)-3-hydroxy-3-methylglutaryl-CoA")
        sub_reactant_2 = abundance("CHEBI", "NADPH")
        sub_reactant_3 = abundance("CHEBI", "hydron")
        sub_product_1 = abundance("CHEBI", "mevalonate")
        sub_product_2 = abundance("CHEBI", "CoA-SH")
        sub_product_3 = abundance("CHEBI", "NADP(+)")

        self.assert_has_node(sub_reactant_1)
        self.assert_has_node(sub_reactant_2)
        self.assert_has_node(sub_reactant_3)
        self.assert_has_node(sub_product_1)
        self.assert_has_node(sub_product_2)
        self.assert_has_node(sub_product_3)

        sub = reaction(
            reactants=[sub_reactant_1, sub_reactant_2, sub_reactant_3],
            products=[sub_product_1, sub_product_2, sub_product_3],
        )
        self.assert_has_node(sub)

        self.assert_has_edge(sub, sub_reactant_1, relation=HAS_REACTANT)
        self.assert_has_edge(sub, sub_reactant_2, relation=HAS_REACTANT)
        self.assert_has_edge(sub, sub_reactant_3, relation=HAS_REACTANT)
        self.assert_has_edge(sub, sub_product_1, relation=HAS_PRODUCT)
        self.assert_has_edge(sub, sub_product_2, relation=HAS_PRODUCT)
        self.assert_has_edge(sub, sub_product_3, relation=HAS_PRODUCT)

    def assert_has_two_way_edge(self, source, target, relation):
        self.assert_has_node(source)
        self.assert_has_node(target)
        self.assert_has_edge(source, target, **{RELATION: relation})
        self.assert_has_edge(target, source, **{RELATION: relation})

    def test_no_correlation(self):
        """Test the ``noCorrelation`` relation."""
        statement = "r(HGNC:X) noCorrelation r(HGNC:Y)"
        self.parser.relation.parseString(statement)
        source = Rna("HGNC", "X")
        target = Rna("HGNC", "Y")
        self.assert_has_two_way_edge(source, target, NO_CORRELATION)

    def test_correlation(self):
        """Test the ``correlation`` relation."""
        statement = "r(HGNC:X) correlation r(HGNC:Y)"
        self.parser.relation.parseString(statement)

        source = Rna("HGNC", "X")
        target = Rna("HGNC", "Y")
        self.assert_has_two_way_edge(source, target, CORRELATION)

    def test_binds(self):
        """Test the ``binds`` relation."""
        statement = "p(HGNC:X) binds p(HGNC:Y)"
        self.parser.relation.parseString(statement)

        source = Protein("HGNC", "X")
        target = Protein("HGNC", "Y")
        x_y_complex = ComplexAbundance([source, target])
        self.assert_has_node(x_y_complex)
        self.assert_has_edge(source, x_y_complex, relation=PART_OF)
        self.assert_has_edge(target, x_y_complex, relation=PART_OF)


class TestCustom(unittest.TestCase):
    def setUp(self):
        graph = BELGraph()

        namespace_to_term = {
            "HGNC": {(None, "AKT1"): "GRP", (None, "YFG"): "GRP"},
            "MESH": {(None, "nucleus"): "A"},
        }

        self.parser = BELParser(graph, namespace_to_term_to_encoding=namespace_to_term, autostreamline=False)

    def test_tloc_undefined_namespace(self):
        s = 'tloc(p(HGNC:AKT1), fromLoc(MESH:nucleus), toLoc(MISSING:"undefined"))'

        with self.assertRaises(UndefinedNamespaceWarning):
            self.parser.translocation.parseString(s)

    def test_tloc_undefined_name(self):
        s = 'tloc(p(HGNC:AKT1), fromLoc(MESH:nucleus), toLoc(MESH:"undefined"))'

        with self.assertRaises(MissingNamespaceNameWarning):
            self.parser.translocation.parseString(s)

    def test_location_undefined_namespace(self):
        s = 'p(HGNC:AKT1, loc(MISSING:"nucleus")'

        with self.assertRaises(UndefinedNamespaceWarning):
            self.parser.protein.parseString(s)

    def test_location_undefined_name(self):
        s = 'p(HGNC:AKT1, loc(MESH:"undefined")'

        with self.assertRaises(MissingNamespaceNameWarning):
            self.parser.protein.parseString(s)



# -*- coding: utf-8 -*-

"""Locations.

Location data also is added into the information in the edge for the node (subject or object) for which it was
annotated. :code:`p(HGNC:GSK3B, pmod(P, S, 9), loc(GO:lysozome)) pos act(p(HGNC:GSK3B), ma(kin))` becomes:

.. code-block:: python

    from pybel.constants import *

    {
        SUBJECT: {
            LOCATION: {
                NAMESPACE: 'GO',
                NAME: 'lysozome',
            }
        },
        RELATION: POSITIVE_CORRELATION,
        OBJECT: {
            MODIFIER: ACTIVITY,
            EFFECT: {
                NAMESPACE: BEL_DEFAULT_NAMESPACE
                NAME: 'kin',
            }
        },
        EVIDENCE: ...,
        CITATION: { ... },
    }


The addition of the :code:`location()` element in BEL 2.0 allows for the unambiguous expression of the differences
between the process of hypothetical :code:`HGNC:A` moving from one place to another and the existence of
hypothetical :code:`HGNC:A` in a specific location having different effects. In BEL 1.0, this action had its own node,
but this introduced unnecessary complexity to the network and made querying more difficult.
This calls for thoughtful consideration of the following two statements:

- :code:`tloc(p(HGNC:A), fromLoc(GO:intracellular), toLoc(GO:"cell membrane")) -> p(HGNC:B)`
- :code:`p(HGNC:A, location(GO:"cell membrane")) -> p(HGNC:B)`

.. seealso::

    - BEL 2.0 specification on `cellular location (2.2.4) <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_cellular_location>`_
    - PyBEL module :py:class:`pybel.parser.modifiers.get_location_language`
"""

from pyparsing import Group, ParserElement, Suppress, oneOf

from ..utils import nest
from ...constants import LOCATION

__all__ = [
    "get_location_language",
]

location_tag = Suppress(oneOf(["loc", "location"]))


def get_location_language(identifier: ParserElement) -> ParserElement:
    """Build a location parser."""
    return Group(location_tag + nest(identifier))(LOCATION)



# -*- coding: utf-8 -*-

"""Tests for data structures in PyBEL."""

import os
import random
import tempfile
import unittest
from io import StringIO

import pybel
import pybel.examples
from pybel import BELGraph
from pybel.constants import CITATION_TYPE_PUBMED, IDENTIFIER, NAMESPACE
from pybel.dsl import hgvs, protein
from pybel.io.api import InvalidExtensionError
from pybel.language import Entity
from pybel.testing.utils import n


class TestGraphProperties(unittest.TestCase):
    """Test setting and access to graph properties."""

    def setUp(self):
        """Make fake metadata for the graphs."""
        (
            self.name,
            self.version,
            self.description,
            self.authors,
            self.contact,
            self.licenses,
            self.copyrights,
            self.disclaimer,
        ) = [n() for _ in range(8)]

    def _help_test_metadata(self, graph: BELGraph) -> None:
        """Help test the right metadata got in the graph."""
        self.assertEqual(self.name, graph.name)
        self.assertEqual(self.version, graph.version)
        self.assertEqual(self.description, graph.description)
        self.assertEqual(self.authors, graph.authors)
        self.assertEqual(self.contact, graph.contact)
        self.assertEqual(self.licenses, graph.license)
        self.assertEqual(self.copyrights, graph.copyright)
        self.assertEqual(self.disclaimer, graph.disclaimer)

        self.assertEqual("{name} v{version}".format(name=self.name, version=self.version), str(graph))

    def test_str_kwargs(self):
        """Test setting of metadata through keyword arguments."""
        graph = BELGraph(
            name=self.name,
            version=self.version,
            description=self.description,
            authors=self.authors,
            contact=self.contact,
            license=self.licenses,
            copyright=self.copyrights,
            disclaimer=self.disclaimer,
        )
        self._help_test_metadata(graph)

    def test_name(self):
        """Test setting of metadata through attributes."""
        graph = BELGraph()

        graph.name = self.name
        graph.version = self.version
        graph.description = self.description
        graph.authors = self.authors
        graph.contact = self.contact
        graph.license = self.licenses
        graph.copyright = self.copyrights
        graph.disclaimer = self.disclaimer

        self._help_test_metadata(graph)


class TestStruct(unittest.TestCase):
    """Test the BEL graph data structure."""

    def test_add_simple(self):
        """Test that a simple node can be added, but not duplicated."""
        graph = BELGraph()
        node = protein(namespace="TEST", name="YFG")
        graph.add_node_from_data(node)
        self.assertEqual(1, graph.number_of_nodes())

        graph.add_node_from_data(node)
        self.assertEqual(1, graph.number_of_nodes(), msg="should not add same node again")

    def test_summarize(self):
        """Test summarizing a graph."""
        self.maxDiff = None
        sio = StringIO()

        random.seed(5)
        pybel.examples.sialic_acid_graph.version = "1.0.0"
        pybel.examples.sialic_acid_graph.summarize(file=sio, examples=False)
        test_str = """---------------------  -------------------
Name                   Sialic Acid Graph
Version                1.0.0
Authors                Charles Tapley Hoyt
Number of Nodes        9
Number of Namespaces   3
Number of Edges        11
Number of Annotations  2
Number of Citations    1
Number of Authors      0
Number of Components   1
Number of Warnings     0
Network Density        1.53E-01
---------------------  -------------------

Type (3)      Count
----------  -------
Protein           7
Complex           1
Abundance         1

Namespace (3)      Count
---------------  -------
go                    15
hgnc                   8
chebi                  2

Edge Type (7)                        Count
---------------------------------  -------
Protein increases Protein                3
Protein directlyIncreases Protein        2
Protein directlyDecreases Protein        2
Complex increases Protein                1
Abundance partOf Complex                 1
Protein partOf Complex                   1
Protein hasVariant Protein               1"""
        self.assertEqual(test_str.strip(), sio.getvalue().strip())

    def test_citation_type_error(self):
        """Test error handling on adding qualified edges."""
        graph = BELGraph()

        with self.assertRaises(TypeError):
            graph.add_increases(
                protein(namespace="TEST", name="YFG1"),
                protein(namespace="TEST", name="YFG2"),
                evidence=n(),
                citation=5,
            )


class TestGetGraphProperties(unittest.TestCase):
    """The tests in this class check the getting and setting of node properties."""

    def setUp(self):
        """Set up the test case with a fresh BEL graph."""
        self.graph = BELGraph()
        self.graph.annotation_pattern["Species"] = r"\d+"
        self.graph.annotation_list["Confidence"] = {
            "Very Low",
            "Low",
            "Medium",
            "High",
            "Very High",
        }

    def test_get_qualified_edge(self):
        """Test adding an edge to a graph."""
        test_source = protein(namespace="TEST", name="YFG")
        test_target = protein(namespace="TEST", name="YFG2")

        self.graph.add_node_from_data(test_source)
        self.graph.add_node_from_data(test_target)

        test_evidence = n()
        test_pmid = n()

        test_key = self.graph.add_increases(
            test_source,
            test_target,
            citation=test_pmid,
            evidence=test_evidence,
            annotations={"Species": "9606", "Confidence": "Very High"},
        )

        citation = self.graph.get_edge_citation(test_source, test_target, test_key)

        self.assertIsNotNone(citation)
        self.assertIsInstance(citation, dict)
        self.assertIn(NAMESPACE, citation)
        self.assertEqual(CITATION_TYPE_PUBMED, citation[NAMESPACE])
        self.assertIn(IDENTIFIER, citation)
        self.assertEqual(test_pmid, citation[IDENTIFIER])

        evidence = self.graph.get_edge_evidence(test_source, test_target, test_key)

        self.assertIsNotNone(evidence)
        self.assertIsInstance(evidence, str)
        self.assertEqual(test_evidence, evidence)

        annotations = self.graph.get_edge_annotations(test_source, test_target, test_key)
        self.assertIsNotNone(annotations)
        self.assertIsInstance(annotations, dict)
        self.assertIn("Species", annotations)
        self.assertIn(Entity(namespace="Species", identifier="9606"), annotations["Species"])
        self.assertIn("Confidence", annotations)
        self.assertIn(
            Entity(namespace="Confidence", identifier="Very High"),
            annotations["Confidence"],
        )

    def test_get_unqualified_edge(self):
        """Test adding an unqualified edge."""
        test_source = protein(namespace="TEST", name="YFG")
        test_target = protein(namespace="TEST", name="YFG2")

        key = self.graph.add_part_of(test_source, test_target)

        citation = self.graph.get_edge_citation(test_source, test_target, key)
        self.assertIsNone(citation)

        evidence = self.graph.get_edge_evidence(test_source, test_target, key)
        self.assertIsNone(evidence)

        annotations = self.graph.get_edge_annotations(test_source, test_target, key)
        self.assertIsNone(annotations)

    def test_add_node_with_variant(self):
        """Test that the identifier is carried through to the child."""
        graph = BELGraph()
        namespace, name, identifier, variant_name = n(), n(), n(), n()
        node = protein(
            namespace=namespace,
            name=name,
            identifier=identifier,
            variants=hgvs(variant_name),
        )
        node.get_parent()

        graph.add_node_from_data(node)

        self.assertEqual(2, graph.number_of_nodes())


class TestExtensionIO(unittest.TestCase):
    def test_io(self):
        with tempfile.TemporaryDirectory() as directory:
            path = os.path.join(directory, "ampk.bel.nodelink.json")
            pybel.dump(pybel.examples.ampk_graph, path)
            self.assertTrue(os.path.exists(path))
            new_graph = pybel.load(path)
            self.assertIsNotNone(new_graph)

    def test_invalid_io(self):
        with tempfile.TemporaryDirectory() as directory:
            path = os.path.join(directory, "ampk.bel.invalid.json")
            with self.assertRaises(InvalidExtensionError):
                pybel.dump(pybel.examples.ampk_graph, path)
            self.assertFalse(os.path.exists(path))



# -*- coding: utf-8 -*-

"""Tests for node predicates."""

import unittest

from pybel import BELGraph
from pybel.constants import (
    ACTIVITY,
    ANNOTATIONS,
    ASSOCIATION,
    CAUSES_NO_CHANGE,
    CITATION,
    CITATION_AUTHORS,
    CITATION_TYPE_PUBMED,
    CITATION_TYPE_URL,
    DECREASES,
    DEGRADATION,
    DIRECTLY_DECREASES,
    DIRECTLY_INCREASES,
    EVIDENCE,
    GMOD,
    IDENTIFIER,
    INCREASES,
    LOCATION,
    MODIFIER,
    NAMESPACE,
    POLAR_RELATIONS,
    POSITIVE_CORRELATION,
    RELATION,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
    TRANSLOCATION,
)
from pybel.dsl import (
    abundance,
    activity,
    degradation,
    fragment,
    gene,
    gmod,
    hgvs,
    pmod,
    protein,
    secretion,
    translocation,
)
from pybel.language import Entity
from pybel.struct.filters import false_node_predicate, true_node_predicate
from pybel.struct.filters.edge_predicate_builders import build_relation_predicate
from pybel.struct.filters.edge_predicates import (
    edge_has_activity,
    edge_has_annotation,
    edge_has_degradation,
    edge_has_translocation,
    has_authors,
    has_polarity,
    has_provenance,
    has_pubmed,
    is_associative_relation,
    is_causal_relation,
    is_direct_causal_relation,
)
from pybel.struct.filters.node_predicates import (
    has_activity,
    has_causal_in_edges,
    has_causal_out_edges,
    has_fragment,
    has_gene_modification,
    has_hgvs,
    has_protein_modification,
    has_variant,
    is_abundance,
    is_causal_central,
    is_causal_sink,
    is_causal_source,
    is_degraded,
    is_gene,
    is_pathology,
    is_protein,
    is_translocated,
    none_of,
    not_pathology,
    one_of,
)
from pybel.testing.utils import n

p1 = protein(name="BRAF", namespace="HGNC")
p2 = protein(name="BRAF", namespace="HGNC", variants=[hgvs("p.Val600Glu"), pmod("Ph")])
p3 = protein(name="APP", namespace="HGNC", variants=fragment(start=672, stop=713))
p4 = protein(name="2", namespace="HGNC")

g1 = gene(name="BRAF", namespace="HGNC", variants=gmod("Me"))


class TestNodePredicates(unittest.TestCase):
    """Tests for node predicates."""

    def test_none_data(self):
        """Test permissive node predicate with a node data dictionary."""
        self.assertTrue(true_node_predicate(p1))
        self.assertFalse(false_node_predicate(p1))

    def test_none(self):
        """Test permissive node predicate with graph and tuple."""
        g = BELGraph()
        g.add_node_from_data(p1)
        self.assertTrue(true_node_predicate(g, p1))
        self.assertFalse(false_node_predicate(g, p1))

    def test_p1_data_variants(self):
        """Test node predicates on BRAF."""
        self.assertFalse(is_abundance(p1))
        self.assertFalse(is_gene(p1))
        self.assertTrue(is_protein(p1))
        self.assertFalse(is_pathology(p1))
        self.assertTrue(not_pathology(p1))

        self.assertFalse(has_variant(p1))
        self.assertFalse(has_protein_modification(p1))
        self.assertFalse(has_gene_modification(p1))
        self.assertFalse(has_hgvs(p1))
        self.assertFalse(has_fragment(p1))

    def test_p1_tuple_variants(self):
        """Test node predicates on the node tuple from BRAF.s"""
        g = BELGraph()
        g.add_node_from_data(p1)

        self.assertFalse(is_abundance(g, p1))
        self.assertFalse(is_gene(g, p1))
        self.assertTrue(is_protein(g, p1))
        self.assertFalse(is_pathology(g, p1))
        self.assertTrue(not_pathology(g, p1))

        self.assertFalse(has_variant(g, p1))
        self.assertFalse(has_protein_modification(g, p1))
        self.assertFalse(has_gene_modification(g, p1))
        self.assertFalse(has_hgvs(g, p1))

    def test_p2_data_variants(self):
        self.assertFalse(is_abundance(p2))
        self.assertFalse(is_gene(p2))
        self.assertTrue(is_protein(p2))
        self.assertFalse(is_pathology(p2))
        self.assertTrue(not_pathology(p2))

        self.assertTrue(has_variant(p2))
        self.assertFalse(has_gene_modification(p2))
        self.assertTrue(has_protein_modification(p2))
        self.assertTrue(has_hgvs(p2))

    def test_p2_tuple_variants(self):
        g = BELGraph()
        g.add_node_from_data(p2)

        self.assertFalse(is_abundance(g, p2))
        self.assertFalse(is_gene(g, p2))
        self.assertTrue(is_protein(g, p2))
        self.assertFalse(is_pathology(g, p2))
        self.assertTrue(not_pathology(g, p2))

        self.assertTrue(has_variant(g, p2))
        self.assertFalse(has_gene_modification(g, p2))
        self.assertTrue(has_protein_modification(g, p2))
        self.assertTrue(has_hgvs(g, p2))

    def test_p3(self):
        self.assertFalse(is_abundance(p3))
        self.assertFalse(is_gene(p3))
        self.assertTrue(is_protein(p3))
        self.assertFalse(is_pathology(p3))
        self.assertTrue(not_pathology(p3))

        self.assertTrue(has_variant(p3))
        self.assertFalse(has_gene_modification(p3))
        self.assertFalse(has_protein_modification(p3))
        self.assertFalse(has_hgvs(p3))
        self.assertTrue(has_fragment(p3))

    def test_g1_variants(self):
        self.assertFalse(is_abundance(g1))
        self.assertTrue(is_gene(g1))
        self.assertFalse(is_protein(g1))
        self.assertFalse(is_pathology(g1))

        self.assertTrue(has_variant(g1))
        self.assertTrue(has_gene_modification(g1), msg="Should have {}: {}".format(GMOD, g1))
        self.assertFalse(has_protein_modification(g1))
        self.assertFalse(has_hgvs(g1))

    def test_fragments(self):
        self.assertTrue(
            has_fragment(
                protein(
                    name="APP",
                    namespace="HGNC",
                    variants=[fragment(start=672, stop=713, description="random text")],
                )
            )
        )
        self.assertTrue(has_fragment(protein(name="APP", namespace="HGNC", variants=[fragment()])))

    def test_p1_active(self):
        """cat(p(HGNC:HSD11B1)) increases deg(a(CHEBI:cortisol))"""
        g = BELGraph()
        g.annotation_pattern["Species"] = r"\d+"

        u = protein(name="HSD11B1", namespace="HGNC")
        v = abundance(name="cortisol", namespace="CHEBI", identifier="17650")

        g.add_increases(
            u,
            v,
            citation={
                NAMESPACE: CITATION_TYPE_URL,
                IDENTIFIER: "https://www.ncbi.nlm.nih.gov/gene/3290",
            },
            evidence="Entrez Gene Summary: Human: The protein encoded by this gene is a microsomal enzyme that "
            "catalyzes the conversion of the stress hormone cortisol to the inactive metabolite cortisone. "
            "In addition, the encoded protein can catalyze the reverse reaction, the conversion of cortisone "
            "to cortisol. Too much cortisol can lead to central obesity, and a particular variation in this "
            "gene has been associated with obesity and insulin resistance in children. Two transcript "
            "variants encoding the same protein have been found for this gene.",
            annotations={"Species": "9606"},
            source_modifier=activity("cat"),
            target_modifier=degradation(),
        )

        self.assertFalse(is_translocated(g, u))
        self.assertFalse(is_degraded(g, u))
        self.assertTrue(has_activity(g, u))

        self.assertFalse(is_translocated(g, v))
        self.assertTrue(is_degraded(g, v))
        self.assertFalse(has_activity(g, v))

    def test_object_has_translocation(self):
        """p(HGNC: EGF) increases tloc(p(HGNC: VCP), GO:0005634, GO:0005737)"""
        g = BELGraph()
        g.annotation_pattern["Species"] = r"\d+"
        u = protein(name="EFG", namespace="HGNC")
        v = protein(name="VCP", namespace="HGNC")

        g.add_increases(
            u,
            v,
            citation="10855792",
            evidence="Although found predominantly in the cytoplasm and, less abundantly, in the nucleus, VCP can be "
            "translocated from the nucleus after stimulation with epidermal growth factor.",
            annotations={"Species": "9606"},
            target_modifier=translocation(
                from_loc=Entity(namespace="GO", identifier="0005634"),
                to_loc=Entity(namespace="GO", identifier="0005737"),
            ),
        )

        self.assertFalse(is_translocated(g, u))
        self.assertFalse(is_degraded(g, u))
        self.assertFalse(has_activity(g, u))
        self.assertFalse(has_causal_in_edges(g, u))
        self.assertTrue(has_causal_out_edges(g, u))

        self.assertTrue(is_translocated(g, v))
        self.assertFalse(is_degraded(g, v))
        self.assertFalse(has_activity(g, v))
        self.assertTrue(has_causal_in_edges(g, v))
        self.assertFalse(has_causal_out_edges(g, v))

    def test_object_has_secretion(self):
        """p(MGI:Il4) increases sec(p(MGI:Cxcl1))"""
        g = BELGraph()
        g.annotation_pattern["Species"] = r"\d+"
        g.annotation_pattern["MeSH"] = ".*"
        u = protein(name="Il4", namespace="MGI")
        v = protein(name="Cxcl1", namespace="MGI")

        g.add_increases(
            u,
            v,
            citation="10072486",
            evidence="Compared with controls treated with culture medium alone, IL-4 and IL-5 induced significantly "
            "higher levels of MIP-2 and KC production; IL-4 also increased the production of MCP-1 "
            "(Fig. 2, A and B)....we only tested the effects of IL-3, IL-4, IL-5, and IL-13 on chemokine "
            "expression and cellular infiltration....Recombinant cytokines were used, ... to treat naive "
            "BALB/c mice.",
            annotations={"Species": "10090", "MeSH": "bronchoalveolar lavage fluid"},
            target_modifier=secretion(),
        )

        self.assertFalse(is_translocated(g, u))
        self.assertFalse(is_degraded(g, u))
        self.assertFalse(has_activity(g, u))
        self.assertFalse(has_causal_in_edges(g, u))
        self.assertTrue(has_causal_out_edges(g, u))

        self.assertTrue(is_translocated(g, v))
        self.assertFalse(is_degraded(g, v))
        self.assertFalse(has_activity(g, v))
        self.assertTrue(has_causal_in_edges(g, v))
        self.assertFalse(has_causal_out_edges(g, v))

    def test_subject_has_secretion(self):
        """sec(p(MGI:S100b)) increases a(CHEBI:"nitric oxide")"""
        g = BELGraph()
        g.annotation_pattern["Species"] = r"\d+"
        g.annotation_pattern["Cell"] = r".*"
        u = protein(name="S100b", namespace="MGI")
        v = abundance(name="nitric oxide", namespace="CHEBI")

        g.add_increases(
            u,
            v,
            citation="11180510",
            evidence="S100B protein is also secreted by astrocytes and acts on these cells to stimulate nitric oxide "
            "secretion in an autocrine manner.",
            annotations={"Species": "10090", "Cell": "astrocyte"},
            source_modifier=secretion(),
        )

        self.assertTrue(is_translocated(g, u))
        self.assertFalse(is_degraded(g, u))
        self.assertFalse(has_activity(g, u))
        self.assertFalse(has_causal_in_edges(g, u))
        self.assertTrue(has_causal_out_edges(g, u))

        self.assertFalse(is_translocated(g, v))
        self.assertFalse(is_degraded(g, v))
        self.assertFalse(has_activity(g, v))
        self.assertTrue(has_causal_in_edges(g, v))
        self.assertFalse(has_causal_out_edges(g, v))

    def test_node_exclusion_data(self):
        g = BELGraph()

        u = protein(name="S100b", namespace="MGI")
        v = abundance(name="nitric oxide", namespace="CHEBI")
        w = abundance(name="cortisol", namespace="CHEBI", identifier="17650")

        g.add_node_from_data(u)
        g.add_node_from_data(v)
        g.add_node_from_data(w)

        f = none_of([u])

        self.assertFalse(f(u))
        self.assertTrue(f(v))
        self.assertTrue(f(w))

        f = none_of([u, v])

        self.assertFalse(f(u))
        self.assertFalse(f(v))
        self.assertTrue(f(w))

        f = none_of([])

        self.assertTrue(f(u))
        self.assertTrue(f(v))
        self.assertTrue(f(w))

    def test_node_exclusion_tuples(self):
        g = BELGraph()

        u = protein(name="S100b", namespace="MGI")
        v = abundance(name="nitric oxide", namespace="CHEBI")
        w = abundance(name="cortisol", namespace="CHEBI", identifier="17650")

        g.add_node_from_data(u)
        g.add_node_from_data(v)
        g.add_node_from_data(w)

        f = none_of([u])

        self.assertFalse(f(g, u))
        self.assertTrue(f(g, v))
        self.assertTrue(f(g, w))

        f = none_of([u, v])

        self.assertFalse(f(g, u))
        self.assertFalse(f(g, v))
        self.assertTrue(f(g, w))

        f = none_of([])

        self.assertTrue(f(g, u))
        self.assertTrue(f(g, v))
        self.assertTrue(f(g, w))

    def test_node_inclusion_data(self):
        g = BELGraph()

        u = protein(name="S100b", namespace="MGI")
        v = abundance(name="nitric oxide", namespace="CHEBI")
        w = abundance(name="cortisol", namespace="CHEBI", identifier="17650")

        g.add_node_from_data(u)
        g.add_node_from_data(v)
        g.add_node_from_data(w)

        f = one_of([u])

        self.assertTrue(f(u))
        self.assertFalse(f(v))
        self.assertFalse(f(w))

        f = one_of([u, v])

        self.assertTrue(f(u))
        self.assertTrue(f(v))
        self.assertFalse(f(w))

        f = one_of([])

        self.assertFalse(f(u))
        self.assertFalse(f(v))
        self.assertFalse(f(w))

    def test_node_inclusion_tuples(self):
        g = BELGraph()

        u = protein(name="S100b", namespace="MGI")
        v = abundance(name="nitric oxide", namespace="CHEBI")
        w = abundance(name="cortisol", namespace="CHEBI", identifier="17650")

        g.add_node_from_data(u)
        g.add_node_from_data(v)
        g.add_node_from_data(w)

        f = one_of([u])

        self.assertTrue(f(g, u))
        self.assertFalse(f(g, v))
        self.assertFalse(f(g, w))

        f = one_of([u, v])

        self.assertTrue(f(g, u))
        self.assertTrue(f(g, v))
        self.assertFalse(f(g, w))

        f = one_of([])

        self.assertFalse(f(g, u))
        self.assertFalse(f(g, v))
        self.assertFalse(f(g, w))

    def test_causal_source(self):
        g = BELGraph()
        a, b, c = (protein(n(), n()) for _ in range(3))

        g.add_increases(a, b, citation=n(), evidence=n())
        g.add_increases(b, c, citation=n(), evidence=n())

        self.assertTrue(is_causal_source(g, a))
        self.assertFalse(is_causal_central(g, a))
        self.assertFalse(is_causal_sink(g, a))

        self.assertFalse(is_causal_source(g, b))
        self.assertTrue(is_causal_central(g, b))
        self.assertFalse(is_causal_sink(g, b))

        self.assertFalse(is_causal_source(g, c))
        self.assertFalse(is_causal_central(g, c))
        self.assertTrue(is_causal_sink(g, c))


class TestEdgePredicate(unittest.TestCase):
    def test_has_polarity_dict(self):
        for relation in POLAR_RELATIONS:
            self.assertTrue(has_polarity({RELATION: relation}))

        self.assertFalse(has_polarity({RELATION: ASSOCIATION}))

    def test_has_polarity(self):
        g = BELGraph()
        a, b, c = (protein(n(), n()) for _ in range(3))
        key1 = g.add_increases(a, b, citation=n(), evidence=n())
        self.assertTrue(has_polarity(g, a, b, key1))

        key2 = g.add_association(b, c, citation=n(), evidence=n())
        self.assertFalse(has_polarity(g, b, c, key2))

    def test_has_provenance(self):
        self.assertFalse(has_provenance({}))
        self.assertFalse(has_provenance({CITATION: {}}))
        self.assertFalse(has_provenance({EVIDENCE: ""}))
        self.assertTrue(has_provenance({CITATION: {}, EVIDENCE: ""}))

    def test_has_pubmed(self):
        self.assertTrue(has_pubmed({CITATION: {NAMESPACE: CITATION_TYPE_PUBMED}}))
        self.assertFalse(has_pubmed({CITATION: {NAMESPACE: CITATION_TYPE_URL}}))
        self.assertFalse(has_pubmed({}))

    def test_has_authors(self):
        self.assertFalse(has_authors({}))
        self.assertFalse(has_authors({CITATION: {}}))
        self.assertFalse(has_authors({CITATION: {CITATION_AUTHORS: []}}))
        self.assertTrue(has_authors({CITATION: {CITATION_AUTHORS: ["One guy"]}}))

    def test_is_causal(self):
        self.assertTrue(is_causal_relation({RELATION: INCREASES}))
        self.assertTrue(is_causal_relation({RELATION: DECREASES}))
        self.assertTrue(is_causal_relation({RELATION: DIRECTLY_INCREASES}))
        self.assertTrue(is_causal_relation({RELATION: DIRECTLY_DECREASES}))

        self.assertFalse(is_causal_relation({RELATION: ASSOCIATION}))
        self.assertFalse(is_causal_relation({RELATION: POSITIVE_CORRELATION}))

    def test_is_direct_causal(self):
        self.assertTrue(is_direct_causal_relation({RELATION: DIRECTLY_INCREASES}))
        self.assertTrue(is_direct_causal_relation({RELATION: DIRECTLY_DECREASES}))

        self.assertFalse(is_direct_causal_relation({RELATION: INCREASES}))
        self.assertFalse(is_direct_causal_relation({RELATION: DECREASES}))
        self.assertFalse(is_direct_causal_relation({RELATION: ASSOCIATION}))
        self.assertFalse(is_direct_causal_relation({RELATION: POSITIVE_CORRELATION}))

    def test_is_association(self):
        self.assertTrue(is_associative_relation({RELATION: ASSOCIATION}))

        self.assertFalse(is_associative_relation({RELATION: INCREASES}))
        self.assertFalse(is_associative_relation({RELATION: CAUSES_NO_CHANGE}))
        self.assertFalse(is_associative_relation({RELATION: DECREASES}))
        self.assertFalse(is_associative_relation({RELATION: DIRECTLY_INCREASES}))
        self.assertFalse(is_associative_relation({RELATION: DIRECTLY_DECREASES}))

    def test_build_is_association(self):
        """Test build_relation_predicate."""
        alternate_is_associative_relation = build_relation_predicate(ASSOCIATION)

        g = BELGraph()
        g.add_edge(p1, p2, key=0, **{RELATION: ASSOCIATION})
        g.add_edge(p2, p3, key=0, **{RELATION: INCREASES})

        self.assertTrue(alternate_is_associative_relation(g, p1, p2, 0))
        self.assertFalse(alternate_is_associative_relation(g, p2, p3, 0))

    def test_build_is_increases_or_decreases(self):
        """Test build_relation_predicate with multiple relations."""
        is_increase_or_decrease = build_relation_predicate([INCREASES, DECREASES])

        g = BELGraph()
        g.add_edge(p1, p2, key=0, **{RELATION: ASSOCIATION})
        g.add_edge(p2, p3, key=0, **{RELATION: INCREASES})
        g.add_edge(p3, p4, key=0, **{RELATION: DECREASES})

        self.assertFalse(is_increase_or_decrease(g, p1, p2, 0))
        self.assertTrue(is_increase_or_decrease(g, p2, p3, 0))
        self.assertTrue(is_increase_or_decrease(g, p3, p4, 0))

    def test_has_degradation(self):
        self.assertTrue(edge_has_degradation({SOURCE_MODIFIER: {MODIFIER: DEGRADATION}}))
        self.assertTrue(edge_has_degradation({TARGET_MODIFIER: {MODIFIER: DEGRADATION}}))

        self.assertFalse(edge_has_degradation({SOURCE_MODIFIER: {MODIFIER: TRANSLOCATION}}))
        self.assertFalse(edge_has_degradation({SOURCE_MODIFIER: {MODIFIER: ACTIVITY}}))
        self.assertFalse(edge_has_degradation({SOURCE_MODIFIER: {LOCATION: None}}))
        self.assertFalse(edge_has_degradation({TARGET_MODIFIER: {MODIFIER: TRANSLOCATION}}))
        self.assertFalse(edge_has_degradation({TARGET_MODIFIER: {MODIFIER: ACTIVITY}}))
        self.assertFalse(edge_has_degradation({TARGET_MODIFIER: {LOCATION: None}}))

    def test_has_translocation(self):
        self.assertTrue(edge_has_translocation({SOURCE_MODIFIER: {MODIFIER: TRANSLOCATION}}))
        self.assertTrue(edge_has_translocation({TARGET_MODIFIER: {MODIFIER: TRANSLOCATION}}))

        self.assertFalse(edge_has_translocation({SOURCE_MODIFIER: {MODIFIER: ACTIVITY}}))
        self.assertFalse(edge_has_translocation({SOURCE_MODIFIER: {LOCATION: None}}))
        self.assertFalse(edge_has_translocation({SOURCE_MODIFIER: {MODIFIER: DEGRADATION}}))
        self.assertFalse(edge_has_translocation({TARGET_MODIFIER: {MODIFIER: ACTIVITY}}))
        self.assertFalse(edge_has_translocation({TARGET_MODIFIER: {LOCATION: None}}))
        self.assertFalse(edge_has_translocation({TARGET_MODIFIER: {MODIFIER: DEGRADATION}}))

    def test_has_activity(self):
        self.assertTrue(edge_has_activity({SOURCE_MODIFIER: {MODIFIER: ACTIVITY}}))
        self.assertTrue(edge_has_activity({TARGET_MODIFIER: {MODIFIER: ACTIVITY}}))

        self.assertFalse(edge_has_activity({SOURCE_MODIFIER: {MODIFIER: TRANSLOCATION}}))
        self.assertFalse(edge_has_activity({TARGET_MODIFIER: {MODIFIER: TRANSLOCATION}}))
        self.assertFalse(edge_has_activity({SOURCE_MODIFIER: {LOCATION: None}}))
        self.assertFalse(edge_has_activity({SOURCE_MODIFIER: {MODIFIER: DEGRADATION}}))
        self.assertFalse(edge_has_activity({TARGET_MODIFIER: {LOCATION: None}}))
        self.assertFalse(edge_has_activity({TARGET_MODIFIER: {MODIFIER: DEGRADATION}}))

    def test_has_annotation(self):
        self.assertFalse(edge_has_annotation({}, "Subgraph"))
        self.assertFalse(edge_has_annotation({ANNOTATIONS: {}}, "Subgraph"))
        self.assertFalse(edge_has_annotation({ANNOTATIONS: {"Subgraph": None}}, "Subgraph"))
        self.assertTrue(edge_has_annotation({ANNOTATIONS: {"Subgraph": "value"}}, "Subgraph"))
        self.assertFalse(edge_has_annotation({ANNOTATIONS: {"Nope": "value"}}, "Subgraph"))



# -*- coding: utf-8 -*-

"""Functions for grouping sub-graphs."""

import logging
from collections import defaultdict
from typing import Mapping, Optional

from ..graph import BELGraph
from ...constants import ANNOTATIONS
from ...language import Entity

__all__ = [
    "get_subgraphs_by_annotation",
]

logger = logging.getLogger(__name__)


def _get_subgraphs_by_annotation_disregard_undefined(graph: BELGraph, annotation: str) -> Mapping[Entity, BELGraph]:
    result = defaultdict(graph.child)

    for source, target, key, data in graph.edges(keys=True, data=True):
        annotation_dict = data.get(ANNOTATIONS)

        if annotation_dict is None:
            continue

        if annotation not in annotation_dict:
            continue

        for entity in annotation_dict[annotation]:
            result[entity].add_edge(source, target, key=key, **data)

    return dict(result)


def _get_subgraphs_by_annotation_keep_undefined(
    graph: BELGraph,
    annotation: str,
    sentinel: Optional[str],
) -> Mapping[Entity, BELGraph]:
    result = defaultdict(graph.child)

    for source, target, key, data in graph.edges(keys=True, data=True):
        annotation_dict = data.get(ANNOTATIONS)

        if annotation_dict is None or annotation not in annotation_dict:
            result[sentinel].add_edge(source, target, key=key, **data)
        else:
            for entity in annotation_dict[annotation]:
                result[entity].add_edge(source, target, key=key, **data)

    return dict(result)


def get_subgraphs_by_annotation(
    graph: BELGraph,
    annotation: str,
    sentinel: Optional[str] = None,
) -> Mapping[Entity, BELGraph]:
    """Stratify the given graph into sub-graphs based on the values for edges' annotations.

    :param graph: A BEL graph
    :param annotation: The annotation to group by
    :param sentinel: The value to stick unannotated edges into. If none, does not keep undefined.
    """
    if sentinel is not None:
        subgraphs = _get_subgraphs_by_annotation_keep_undefined(graph, annotation, sentinel)
    else:
        subgraphs = _get_subgraphs_by_annotation_disregard_undefined(graph, annotation)

    return subgraphs



# -*- coding: utf-8 -*-

"""Exceptions for the internal DSL."""

from ..exceptions import PyBELWarning

__all__ = [
    "PyBELDSLException",
    "InferCentralDogmaException",
    "ListAbundanceEmptyException",
    "ReactionEmptyException",
]


class PyBELDSLException(PyBELWarning, ValueError):
    """Raised when problems with the DSL."""


class InferCentralDogmaException(PyBELDSLException):
    """Raised when unable to infer central dogma."""


class ListAbundanceEmptyException(PyBELDSLException):
    """Raised when a list abundance has no members."""


class ReactionEmptyException(PyBELDSLException):
    """Raised when a reaction has neither reactants nor products."""



# -*- coding: utf-8 -*-

"""This module contains the base class for connection managers in SQLAlchemy."""

import logging
from typing import List, Optional, Tuple, Type, TypeVar

import pystow
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from .models import Base

__all__ = [
    "BaseManager",
    "build_engine_session",
]

logger = logging.getLogger(__name__)

X = TypeVar("X")


def build_engine_session(
    connection: str,
    echo: bool = False,
    autoflush: Optional[bool] = None,
    autocommit: Optional[bool] = None,
    expire_on_commit: Optional[bool] = None,
    scopefunc=None,
) -> Tuple:
    """Build an engine and a session.

    :param connection: An RFC-1738 database connection string
    :param echo: Turn on echoing SQL
    :param autoflush: Defaults to True if not specified in kwargs or configuration.
    :param autocommit: Defaults to False if not specified in kwargs or configuration.
    :param expire_on_commit: Defaults to False if not specified in kwargs or configuration.
    :param scopefunc: Scoped function to pass to :func:`sqlalchemy.orm.scoped_session`
    :rtype: tuple[Engine,Session]

    From the Flask-SQLAlchemy documentation:

    An extra key ``'scopefunc'`` can be set on the ``options`` dict to
    specify a custom scope function.  If it's not provided, Flask's app
    context stack identity is used. This will ensure that sessions are
    created and removed with the request/response cycle, and should be fine
    in most cases.
    """
    if connection is None:
        raise ValueError("can not build engine when connection is None")

    engine = create_engine(connection, echo=echo)

    autoflush = pystow.get_config("pybel", "manager_autoflush", passthrough=autoflush, dtype=bool, default=True)
    autocommit = pystow.get_config("pybel", "manager_autocommit", passthrough=autocommit, dtype=bool, default=False)
    expire_on_commit = pystow.get_config(
        "pybel",
        "manager_autoexpire",
        passthrough=expire_on_commit,
        dtype=bool,
        default=True,
    )

    logger.debug(
        "auto flush: %s, auto commit: %s, expire on commmit: %s",
        autoflush,
        autocommit,
        expire_on_commit,
    )

    #: A SQLAlchemy session maker
    session_maker = sessionmaker(
        bind=engine,
        autoflush=autoflush,
        autocommit=autocommit,
        expire_on_commit=expire_on_commit,
    )

    #: A SQLAlchemy session object
    session = scoped_session(
        session_maker,
        scopefunc=scopefunc,
    )

    return engine, session


class BaseManager(object):
    """A wrapper around a SQLAlchemy engine and session."""

    #: The declarative base for this manager
    base = Base

    def __init__(self, engine, session) -> None:
        """Instantiate a manager from an engine and session."""
        self.engine = engine
        self.session = session

    def create_all(self, checkfirst: bool = True) -> None:
        """Create the PyBEL cache's database and tables.

        :param checkfirst: Check if the database exists before trying to re-make it
        """
        self.base.metadata.create_all(bind=self.engine, checkfirst=checkfirst)

    def drop_all(self, checkfirst: bool = True) -> None:
        """Drop all data, tables, and databases for the PyBEL cache.

        :param checkfirst: Check if the database exists before trying to drop it
        """
        self.session.close()
        self.base.metadata.drop_all(bind=self.engine, checkfirst=checkfirst)

    def bind(self) -> None:
        """Bind the metadata to the engine and session."""
        self.base.metadata.bind = self.engine
        self.base.query = self.session.query_property()

    def _list_model(self, model_cls: Type[X]) -> List[X]:
        """List the models in this class."""
        return self.session.query(model_cls).all()

    def _count_model(self, model_cls) -> int:
        """Count the number of models in the database."""
        return self.session.query(model_cls).count()

    def __repr__(self):
        return "<{} connection={}>".format(self.__class__.__name__, self.engine.url)



# -*- coding: utf-8 -*-

"""Functions for building graphs that use both expansion and induction procedures."""

import logging
from typing import Iterable, Union

from .expansion import expand_all_node_neighborhoods
from .expansion.upstream import expand_downstream_causal, expand_upstream_causal
from .induction.neighborhood import get_subgraph_by_neighborhood
from .induction.upstream import (
    get_downstream_causal_subgraph,
    get_upstream_causal_subgraph,
)
from ..pipeline import transformation
from ...dsl import BaseEntity

__all__ = [
    "get_multi_causal_upstream",
    "get_multi_causal_downstream",
    "get_subgraph_by_second_neighbors",
]

logger = logging.getLogger(__name__)


@transformation
def get_multi_causal_upstream(graph, nbunch: Union[BaseEntity, Iterable[BaseEntity]]):
    """Get the union of all the 2-level deep causal upstream subgraphs from the nbunch.

    :param pybel.BELGraph graph: A BEL graph
    :param nbunch: A BEL node or list of BEL nodes
    :return: A subgraph of the original BEL graph
    :rtype: pybel.BELGraph
    """
    result = get_upstream_causal_subgraph(graph, nbunch)
    expand_upstream_causal(graph, result)
    return result


@transformation
def get_multi_causal_downstream(graph, nbunch: Union[BaseEntity, Iterable[BaseEntity]]):
    """Get the union of all of the 2-level deep causal downstream subgraphs from the nbunch.

    :param pybel.BELGraph graph: A BEL graph
    :param nbunch: A BEL node or list of BEL nodes
    :return: A subgraph of the original BEL graph
    :rtype: pybel.BELGraph
    """
    result = get_downstream_causal_subgraph(graph, nbunch)
    expand_downstream_causal(graph, result)
    return result


@transformation
def get_subgraph_by_second_neighbors(graph, nodes: Iterable[BaseEntity], filter_pathologies: bool = False):
    """Get a graph around the neighborhoods of the given nodes and expand to the neighborhood of those nodes.

    Returns none if none of the nodes are in the graph.

    :param pybel.BELGraph graph: A BEL graph
    :param nodes: An iterable of BEL nodes
    :param filter_pathologies: Should expansion take place around pathologies?
    :return: A BEL graph induced around the neighborhoods of the given nodes
    :rtype: Optional[pybel.BELGraph]
    """
    result = get_subgraph_by_neighborhood(graph, nodes)

    if result is None:
        return

    expand_all_node_neighborhoods(graph, result, filter_pathologies=filter_pathologies)
    return result



# -*- coding: utf-8 -*-

"""Utilities for BEL graphs."""

import logging
import random
from collections import Counter
from typing import Optional, TextIO

import bioregistry
import pandas as pd
from humanize import intword
from tabulate import tabulate

from .node_summary import count_namespaces
from ..graph import BELGraph
from ...constants import CITATION, IDENTIFIER, NAMESPACE, RELATION, TWO_WAY_RELATIONS
from ...dsl import BaseConcept
from ...utils import multidict

logger = logging.getLogger(__name__)


def function_table_df(graph: BELGraph, examples: bool = True) -> pd.DataFrame:
    """Create a dataframe describing the functions in the graph."""
    function_mapping = multidict((node.function, node) for node in graph)
    function_c = Counter({function: len(nodes) for function, nodes in function_mapping.items()})
    if not examples:
        return pd.DataFrame(function_c.most_common(), columns=["Type", "Count"])
    return pd.DataFrame(
        [
            (function, count, random.choice(function_mapping[function]))  # noqa:S311
            for function, count in function_c.most_common()
        ],
        columns=["Type", "Count", "Example"],
    )


def functions_str(graph, examples: bool = True, add_count: bool = True, **kwargs) -> str:
    """Make a summary string of the functions in the graph."""
    df = function_table_df(graph, examples=examples)
    headers = list(df.columns)
    if add_count:
        headers[0] += " ({})".format(len(df.index))
    return tabulate(df.values, headers=headers, **kwargs)


def functions(graph, file: Optional[TextIO] = None, examples: bool = True, **kwargs) -> None:
    """Print a summary of the functions in the graph."""
    print(functions_str(graph=graph, examples=examples, **kwargs), file=file)


def namespaces_table_df(graph: BELGraph, examples: bool = True) -> pd.DataFrame:
    """Create a dataframe describing the namespaces in the graph."""
    namespace_mapping = multidict((node.namespace, node) for node in graph if isinstance(node, BaseConcept))
    namespace_c = count_namespaces(graph)
    if not examples:
        return pd.DataFrame(namespace_c.most_common(), columns=["Namespace", "Count"])
    return pd.DataFrame(
        [
            (
                prefix,
                bioregistry.get_name(prefix),
                count,
                random.choice(namespace_mapping[prefix]) if prefix in namespace_mapping else "",  # noqa:S311
            )
            for prefix, count in namespace_c.most_common()
        ],
        columns=["Prefix", "Name", "Count", "Example"],
    )


def namespaces_str(graph: BELGraph, examples: bool = True, add_count: bool = True, **kwargs) -> None:
    """Make a summary string of the namespaces in the graph."""
    df = namespaces_table_df(graph, examples=examples)
    headers = list(df.columns)
    if add_count:
        headers[0] += " ({})".format(len(df.index))
    return tabulate(df.values, headers=headers, **kwargs)


def namespaces(graph: BELGraph, file: Optional[TextIO] = None, examples: bool = True, **kwargs) -> None:
    """Print a summary of the namespaces in the graph."""
    print(namespaces_str(graph=graph, examples=examples, **kwargs), file=file)


def edge_table_df(graph: BELGraph, *, examples: bool = True, minimum: Optional[int] = None) -> pd.DataFrame:
    """Create a dataframe describing the edges in the graph."""
    edge_mapping = multidict(
        (
            f"{u.function} {d[RELATION]} {v.function}",
            graph.edge_to_bel(u, v, d, use_identifiers=True),
        )
        for u, v, d in graph.edges(data=True)
        if d[RELATION] not in TWO_WAY_RELATIONS or u.function > v.function
    )
    edge_c = Counter({top_level_edge: len(edges) for top_level_edge, edges in edge_mapping.items()})
    if examples:
        rows = [
            (
                top_level_edge,
                count,
                random.choice(edge_mapping[top_level_edge]),
            )  # noqa:S311
            for top_level_edge, count in edge_c.most_common()
            if not minimum or count >= minimum
        ]
        columns = ["Edge Type", "Count", "Example"]
    else:
        rows = edge_c.most_common()
        if minimum:
            rows = [(k, count) for k, count in rows if count >= minimum]
        columns = ["Edge Type", "Count"]
    return pd.DataFrame(rows, columns=columns)


def edges_str(
    graph: BELGraph,
    *,
    examples: bool = True,
    add_count: bool = True,
    minimum: Optional[int] = None,
    **kwargs,
) -> str:
    """Make a summary str of the edges in the graph."""
    df = edge_table_df(graph, examples=examples, minimum=minimum)
    headers = list(df.columns)
    if add_count:
        headers[0] += " ({})".format(intword(len(df.index)))
    return tabulate(df.values, headers=headers, **kwargs)


def edges(
    graph: BELGraph,
    *,
    examples: bool = True,
    minimum: Optional[int] = None,
    file: Optional[TextIO] = None,
    **kwargs,
) -> None:
    """Print a summary of the edges in the graph."""
    print(edges_str(graph=graph, examples=examples, minimum=minimum, **kwargs), file=file)


def citations(graph: BELGraph, n: Optional[int] = 15, file: Optional[TextIO] = None) -> None:
    """Print a summary of the citations in the graph."""
    edge_mapping = multidict(
        (
            (data[CITATION][NAMESPACE], data[CITATION][IDENTIFIER]),
            graph.edge_to_bel(u, v, data),
        )
        for u, v, data in graph.edges(data=True)
        if CITATION in data
    )
    edge_c = Counter({top_level_edge: len(edges) for top_level_edge, edges in edge_mapping.items()})
    df = pd.DataFrame(
        [
            (
                ":".join(top_level_edge),
                count,
                random.choice(edge_mapping[top_level_edge]),
            )  # noqa:S311
            for top_level_edge, count in edge_c.most_common(n=n)
        ],
        columns=["Citation", "Count", "Example"],
    )

    if n is None or len(edge_mapping) < n:
        print("{} Citation Count: {}".format(graph, len(edge_mapping)))
    else:
        print("{} Citation Count: {} (Showing top {})".format(graph, len(edge_mapping), n))
    print(tabulate(df.values, headers=df.columns), file=file)



# -*- coding: utf-8 -*-

"""Utility functions for grouping sub-graphs by citation."""

from collections import defaultdict
from typing import Mapping, Tuple

from ..graph import BELGraph
from ...constants import CITATION, IDENTIFIER, NAMESPACE

__all__ = [
    "get_subgraphs_by_citation",
]


def get_subgraphs_by_citation(graph: BELGraph) -> Mapping[Tuple[str, str], BELGraph]:
    """Stratify the graph based on citations.

    :param graph: A BEL graph
    :return: A mapping of each citation db/id to the BEL graph from it.
    """
    rv = defaultdict(graph.child)

    for u, v, key, data in graph.edges(keys=True, data=True):
        if CITATION not in data:
            continue
        dk = data[CITATION][NAMESPACE], data[CITATION][IDENTIFIER]

        rv[dk].add_edge(u, v, key=key, **data)

    return dict(rv)



# -*- coding: utf-8 -*-

"""This module contains output functions to BEL scripts."""

import gzip
import itertools as itt
import logging
import time
import warnings
from typing import Iterable, List, Mapping, Optional, TextIO, Tuple, Union

import bel_resources.constants
from bel_resources import make_knowledge_header
from networkx.utils import open_file

from .constants import (
    ACTIVITY,
    ANNOTATIONS,
    CITATION,
    CITATION_TYPE_PUBMED,
    DEGRADATION,
    EFFECT,
    EVIDENCE,
    FROM_LOC,
    LOCATION,
    MODIFIER,
    NAME,
    NAMESPACE,
    PYBEL_AUTOEVIDENCE,
    PYBEL_PUBMED,
    RELATION,
    SET_CITATION_FMT,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
    TO_LOC,
    TRANSLOCATION,
    UNQUALIFIED_EDGES,
    VARIANTS,
)
from .dsl import BaseAbundance, BaseEntity, FusionBase, ListAbundance, Reaction
from .language import Entity
from .typing import EdgeData
from .utils import ensure_quotes
from .version import VERSION

__all__ = [
    "to_bel_script",
    "to_bel_script_gz",
    "to_bel_script_lines",
    "edge_to_bel",
    "edge_to_tuple",
    "calculate_canonical_name",
]

logger = logging.getLogger(__name__)

EdgeTuple = Tuple[BaseEntity, BaseEntity, str, EdgeData]


@open_file(1, mode="w")
def to_bel_script(graph, path: Union[str, TextIO], use_identifiers: bool = True) -> None:
    """Write the BELGraph as a canonical BEL script.

    :param BELGraph graph: the BEL Graph to output as a BEL Script
    :param path: A path or file-like.
    :param use_identifiers: Enables extended `BEP-0008 <http://bep.bel.bio/published/BEP-0008.html>`_ syntax
    """
    for line in to_bel_script_lines(graph, use_identifiers=use_identifiers):
        print(line, file=path)


def to_bel_script_gz(graph, path, **kwargs) -> None:
    """Write the graph as a BEL Script a gzip file."""
    with gzip.open(path, "wt") as file:
        to_bel_script(graph, file, **kwargs)


def to_bel_script_lines(graph, use_identifiers: bool = True) -> Iterable[str]:
    """Iterate over the lines of the BEL graph as a canonical BEL script.

    :param pybel.BELGraph graph: A BEL Graph
    :param use_identifiers: Enables extended `BEP-0008 <http://bep.bel.bio/published/BEP-0008.html>`_ syntax
    """
    return itt.chain(
        _to_bel_lines_header(graph),
        _to_bel_lines_body(graph, use_identifiers=use_identifiers),
        _to_bel_lines_footer(graph, use_identifiers=use_identifiers),
    )


def postpend_location(bel_string: str, location_model) -> str:
    """Rip off the closing parentheses and adds canonicalized modification.

    I did this because writing a whole new parsing model for the data would be sad and difficult

    :param bel_string: BEL string representing node
    :param dict location_model: A dictionary containing keys :code:`pybel.constants.TO_LOC` and
                                :code:`pybel.constants.FROM_LOC`
    :return: A part of a BEL string representing the location
    """
    if not all(k in location_model for k in {NAMESPACE, NAME}):
        raise ValueError("Location model missing namespace and/or name keys: {}".format(location_model))

    return "{}, loc({}:{}))".format(
        bel_string[:-1],
        location_model[NAMESPACE],
        ensure_quotes(location_model[NAME]),
    )


def _decanonicalize_edge_node(
    node: BaseEntity,
    edge_data: EdgeData,
    node_position: str,
    *,
    use_identifiers: bool = True,
) -> str:
    """Canonicalize a node with its modifiers stored in the given edge to a BEL string.

    :param node: A PyBEL node data dictionary
    :param edge_data: A PyBEL edge data dictionary
    :param node_position: Either :data:`pybel.constants.SUBJECT` or :data:`pybel.constants.OBJECT`
    :param use_identifiers: Enables extended `BEP-0008 <http://bep.bel.bio/published/BEP-0008.html>`_ syntax
    """
    node_str = node.as_bel(use_identifiers=use_identifiers)

    if node_position not in edge_data:
        return node_str

    node_edge_data = edge_data[node_position]

    if LOCATION in node_edge_data:
        node_str = postpend_location(node_str, node_edge_data[LOCATION])

    modifier = node_edge_data.get(MODIFIER)

    if modifier is None:
        return node_str

    if DEGRADATION == modifier:
        warnings.warn("degradation is deprecated", DeprecationWarning)
        return f"deg({node_str})"

    effect = node_edge_data.get(EFFECT)

    if ACTIVITY == modifier:
        if effect is None:
            return f"act({node_str})"
        return f"act({node_str}, ma({effect}))"

    if TRANSLOCATION == modifier:
        if effect is None:
            return f"tloc({node_str})"

        to_loc_data: Entity = effect[TO_LOC]
        from_loc_data: Entity = effect[FROM_LOC]
        return f"tloc({node_str}, fromLoc({from_loc_data}), toLoc({to_loc_data}))"

    raise ValueError("invalid modifier: {}".format(modifier))


def edge_to_tuple(
    source: BaseEntity,
    target: BaseEntity,
    data: EdgeData,
    use_identifiers: bool = True,
) -> Tuple[str, str, str]:
    """Take two nodes and gives back a BEL string representing the statement.

    :param source: The edge's source's PyBEL node data dictionary
    :param target: The edge's target's PyBEL node data dictionary
    :param data: The edge's data dictionary
    :param use_identifiers: Enables extended `BEP-0008 <http://bep.bel.bio/published/BEP-0008.html>`_ syntax
    """
    u_str = _decanonicalize_edge_node(source, data, node_position=SOURCE_MODIFIER, use_identifiers=use_identifiers)
    v_str = _decanonicalize_edge_node(target, data, node_position=TARGET_MODIFIER, use_identifiers=use_identifiers)
    return u_str, data[RELATION], v_str


def edge_to_bel(
    source: BaseEntity,
    target: BaseEntity,
    data: EdgeData,
    sep: Optional[str] = None,
    use_identifiers: bool = True,
) -> str:
    """Take two nodes and gives back a BEL string representing the statement.

    :param source: The edge's source's PyBEL node data dictionary
    :param target: The edge's target's PyBEL node data dictionary
    :param data: The edge's data dictionary
    :param sep: The separator between the source, relation, and target. Defaults to ' '
    :param use_identifiers: Enables extended `BEP-0008 <http://bep.bel.bio/published/BEP-0008.html>`_ syntax
    """
    sep = sep or " "
    return sep.join(edge_to_tuple(source=source, target=target, data=data, use_identifiers=use_identifiers))


def _sort_qualified_edges_helper(t: EdgeTuple) -> Tuple[str, str, str]:
    return (
        t[3][CITATION].namespace,
        t[3][CITATION].identifier,
        t[3][EVIDENCE],
    )


def sort_qualified_edges(graph) -> Iterable[EdgeTuple]:
    """Return the qualified edges, sorted first by citation, then by evidence, then by annotations.

    :param BELGraph graph: A BEL graph
    """
    qualified_edges = (
        (u, v, k, d)
        for u, v, k, d in graph.edges(keys=True, data=True)
        if graph.has_edge_citation(u, v, k) and graph.has_edge_evidence(u, v, k)
    )
    return sorted(qualified_edges, key=_sort_qualified_edges_helper)


def _citation_sort_key(t: EdgeTuple) -> Tuple[str, str]:
    """Make a confusing 4 tuple sortable by citation."""
    return t[3][CITATION].namespace, t[3][CITATION].identifier


def _evidence_sort_key(t: EdgeTuple) -> str:
    """Make a confusing 4 tuple sortable by citation."""
    return t[3][EVIDENCE]


def _set_annotation_to_str(annotation_data: Mapping[str, List[Entity]], key: str, use_curie: bool = False) -> str:
    """Return a set annotation string."""
    value = annotation_data[key]

    if len(value) == 1:
        value = list(value)[0]
        return f'SET {key} = "{value if use_curie else value.identifier}"'

    value_strings = ", ".join(
        f'"{v if use_curie else v.identifier}"'
        for v in sorted(value, key=lambda e: (e.namespace, e.identifier, e.name))
    )
    return f"SET {key} = {{{value_strings}}}"


def _unset_annotation_to_str(keys: List[str]) -> str:
    """Return an unset annotation string."""
    if len(keys) == 1:
        return "UNSET {}".format(list(keys)[0])

    return "UNSET {{{}}}".format(", ".join("{}".format(key) for key in keys))


def _to_bel_lines_header(graph) -> Iterable[str]:
    """Iterate the lines of a BEL graph's corresponding BEL script's header.

    :param pybel.BELGraph graph: A BEL graph
    """
    yield "# This document was created by PyBEL v{} and bel-resources v{} on {}\n".format(
        VERSION,
        bel_resources.constants.VERSION,
        time.asctime(),
    )
    yield from make_knowledge_header(
        namespace_url=graph.namespace_url,
        namespace_patterns=graph.namespace_pattern,
        annotation_url=graph.annotation_url,
        annotation_patterns=graph.annotation_pattern,
        annotation_list=graph.annotation_list,
        **graph.document,
    )


def group_citation_edges(
    edges: Iterable[EdgeTuple],
) -> Iterable[Tuple[Tuple[str, str], Iterable[EdgeTuple]]]:
    """Return an iterator over pairs of citation values and their corresponding edge iterators."""
    return itt.groupby(edges, key=_citation_sort_key)


def group_evidence_edges(
    edges: Iterable[EdgeTuple],
) -> Iterable[Tuple[str, Iterable[EdgeTuple]]]:
    """Return an iterator over pairs of evidence values and their corresponding edge iterators."""
    return itt.groupby(edges, key=_evidence_sort_key)


def _to_bel_lines_body(graph, use_identifiers: bool = False) -> Iterable[str]:
    """Iterate the lines of a BEL graph's corresponding BEL script's body.

    :param pybel.BELGraph graph: A BEL graph
    :param use_identifiers: Enables extended `BEP-0008 <http://bep.bel.bio/published/BEP-0008.html>`_ syntax
    """
    qualified_edges = sort_qualified_edges(graph)

    for (citation_db, citation_id), citation_edges in group_citation_edges(qualified_edges):
        yield SET_CITATION_FMT.format(citation_db, citation_id) + "\n"

        for evidence, evidence_edges in group_evidence_edges(citation_edges):
            yield 'SET SupportingText = "{}"'.format(evidence)

            for u, v, _, data in evidence_edges:
                annotations_data = data.get(ANNOTATIONS)

                keys = sorted(annotations_data) if annotations_data is not None else tuple()
                for key in keys:
                    yield _set_annotation_to_str(annotations_data, key, use_curie=key in graph.annotation_curie)

                yield graph.edge_to_bel(u, v, data, use_identifiers=use_identifiers)

                if keys:
                    yield _unset_annotation_to_str(keys)

            yield "UNSET SupportingText"
        yield "UNSET Citation\n"
        yield "#" * 80


def _to_bel_lines_footer(graph, use_identifiers: bool = False) -> Iterable[str]:
    """Iterate the lines of a BEL graph's corresponding BEL script's footer.

    :param pybel.BELGraph graph: A BEL graph
    :param use_identifiers: Enables extended `BEP-0008 <http://bep.bel.bio/published/BEP-0008.html>`_ syntax
    """
    unqualified_edges_to_serialize = [
        (u, v, d) for u, v, d in graph.edges(data=True) if d[RELATION] in UNQUALIFIED_EDGES and EVIDENCE not in d
    ]

    isolated_nodes_to_serialize = [node for node in graph if not graph.pred[node] and not graph.succ[node]]

    if unqualified_edges_to_serialize or isolated_nodes_to_serialize:
        yield "###############################################\n"
        yield SET_CITATION_FMT.format(CITATION_TYPE_PUBMED, PYBEL_PUBMED)
        yield 'SET SupportingText = "{}"'.format(PYBEL_AUTOEVIDENCE)

        for u, v, data in unqualified_edges_to_serialize:
            yield "{} {} {}".format(
                u.as_bel(use_identifiers=use_identifiers),
                data[RELATION],
                v.as_bel(use_identifiers=use_identifiers),
            )

        for node in isolated_nodes_to_serialize:
            yield node.as_bel(use_identifiers=use_identifiers)

        yield "UNSET SupportingText"
        yield "UNSET Citation"


def calculate_canonical_name(node: BaseEntity, use_identifiers: bool = True) -> str:
    """Calculate the canonical name for a given node.

    If it is a simple node, uses the already given name. Otherwise, it uses the BEL string.
    """
    if isinstance(node, (Reaction, ListAbundance, FusionBase)):
        return node.as_bel(use_identifiers=True)

    elif isinstance(node, BaseAbundance):
        if VARIANTS in node:
            return node.as_bel(use_identifiers=True)
        elif use_identifiers and node.entity.identifier and node.entity.name:
            return node.obo
        else:
            return node.curie

    else:
        raise TypeError("Unhandled node: {}".format(node))



# -*- coding: utf-8 -*-

"""Functions for deleting nodes and edges in networks."""

from ...filters.edge_filters import filter_edges
from ...filters.edge_predicates import is_associative_relation, not_causal_relation
from ...filters.node_filters import filter_nodes
from ...filters.node_predicates import (
    is_biological_process,
    is_isolated_list_abundance,
    is_pathology,
)
from ...pipeline import in_place_transformation

__all__ = [
    "remove_filtered_edges",
    "remove_filtered_nodes",
    "remove_associations",
    "remove_pathologies",
    "remove_biological_processes",
    "remove_isolated_list_abundances",
    "remove_non_causal_edges",
]


@in_place_transformation
def remove_filtered_edges(graph, edge_predicates=None):
    """Remove edges passing the given edge predicates.

    :param pybel.BELGraph graph: A BEL graph
    :param edge_predicates: A predicate or list of predicates
    :type edge_predicates: None or ((pybel.BELGraph, tuple, tuple, int) -> bool) or iter[(pybel.BELGraph, tuple, tuple, int) -> bool]]
    :return:
    """
    edges = list(filter_edges(graph, edge_predicates=edge_predicates))
    graph.remove_edges_from(edges)


@in_place_transformation
def remove_filtered_nodes(graph, node_predicates=None):
    """Remove nodes passing the given node predicates.

    :param pybel.BELGraph graph: A BEL graph
    :type node_predicates: None or ((pybel.BELGraph, tuple) -> bool) or iter[(pybel.BELGraph, tuple) -> bool)]
    """
    nodes = list(filter_nodes(graph, node_predicates=node_predicates))
    graph.remove_nodes_from(nodes)


@in_place_transformation
def remove_associations(graph):
    """Remove all associative relationships from the graph.

    :param pybel.BELGraph graph: A BEL graph
    """
    remove_filtered_edges(graph, is_associative_relation)


@in_place_transformation
def remove_pathologies(graph):
    """Remove pathology nodes from the graph.

    :param pybel.BELGraph graph: A BEL graph
    """
    remove_filtered_nodes(graph, node_predicates=is_pathology)


@in_place_transformation
def remove_biological_processes(graph):
    """Remove biological process nodes from the graph.

    :param pybel.BELGraph graph: A BEL graph
    """
    remove_filtered_nodes(graph, node_predicates=is_biological_process)


@in_place_transformation
def remove_isolated_list_abundances(graph):
    """Remove isolated list abundances from the graph.

    :param pybel.BELGraph graph: A BEL graph
    """
    remove_filtered_nodes(graph, is_isolated_list_abundance)


@in_place_transformation
def remove_non_causal_edges(graph):
    """Remove non-causal edges from the graph."""
    remove_filtered_edges(graph, not_causal_relation)



# -*- coding: utf-8 -*-

"""Utilities for PyBEL testing."""

from uuid import uuid4

from requests.compat import urlparse

from ..manager import Manager
from ..manager.models import Namespace, NamespaceEntry
from ..struct import BELGraph
from ..struct.summary import (
    get_annotation_values_by_annotation,
    iter_annotation_value_pairs,
)
from ..struct.summary.node_summary import get_names

_FRAUNHOFER_RESOURCES = "https://owncloud.scai.fraunhofer.de/index.php/s/JsfpQvkdx3Y5EMx/download?path="


def get_uri_name(url: str) -> str:
    """Get the file name from the end of the URL."""
    url_parsed = urlparse(url)
    if url.startswith(_FRAUNHOFER_RESOURCES):
        return url_parsed.query.split("=")[-1]
    else:
        url_parts = url_parsed.path.split("/")
        return url_parts[-1]


def n() -> str:
    """Return a UUID string for testing."""
    return str(uuid4())[:15]


def make_dummy_namespaces(manager: Manager, graph: BELGraph) -> None:
    """Make dummy namespaces for the test."""
    for keyword, names in get_names(graph).items():
        graph.namespace_url[keyword] = url = n()

        namespace = Namespace(keyword=keyword, url=url)
        manager.session.add(namespace)

        for name in names:
            entry = NamespaceEntry(name=name, namespace=namespace)
            manager.session.add(entry)

    manager.session.commit()


def make_dummy_annotations(manager: Manager, graph: BELGraph):
    """Make dummy annotations for the test."""
    namespaces = {}
    for keyword, entity in iter_annotation_value_pairs(graph):
        namespace = namespaces.get(keyword)
        if namespace is None:
            graph.annotation_url[keyword] = url = n()
            namespace = Namespace(keyword=keyword, url=url, is_annotation=True)
            manager.session.add(namespace)

        entry = NamespaceEntry(name=entity.name, identifier=entity.identifier, namespace=namespace)
        manager.session.add(entry)

    manager.session.commit()



# -*- coding: utf-8 -*-

"""Tests for transfer of knowledge and inference functions."""

import unittest

from pybel.examples.statin_example import (
    avorastatin,
    ec_11134,
    ec_11188,
    fluvastatin,
    hmgcr,
    hmgcr_inhibitor,
    mevinolinic_acid,
    statin,
    statin_graph,
    synthetic_statin,
)
from pybel.struct.mutation import infer_child_relations
from pybel.struct.mutation.inference.transfer import iter_children


class TestTransfer(unittest.TestCase):
    """Tests for transfer of knowledge and inference functions."""

    def test_get_children(self):
        """Test iterating over the children of a node."""
        children = list(iter_children(statin_graph, hmgcr_inhibitor))

        self.assertNotEqual(0, len(children), msg="no children found")
        self.assertIn(mevinolinic_acid, children, msg="direct child not found")

    def test_infer(self):
        """Test inferring child relations."""
        graph = statin_graph.copy()
        self.assertEqual(9, graph.number_of_nodes())
        self.assertEqual(8, graph.number_of_edges())

        self.assertNotIn(ec_11134, graph[fluvastatin])
        self.assertNotIn(ec_11188, graph[fluvastatin])
        self.assertNotIn(ec_11134, graph[avorastatin])
        self.assertNotIn(ec_11188, graph[avorastatin])
        self.assertNotIn(ec_11134, graph[synthetic_statin])
        self.assertNotIn(ec_11188, graph[synthetic_statin])
        self.assertNotIn(ec_11134, graph[statin])
        self.assertNotIn(ec_11188, graph[statin])
        self.assertNotIn(ec_11134, graph[mevinolinic_acid])
        self.assertNotIn(ec_11188, graph[mevinolinic_acid])
        self.assertIn(ec_11134, graph[hmgcr_inhibitor])
        self.assertIn(ec_11188, graph[hmgcr_inhibitor])

        infer_child_relations(graph, hmgcr_inhibitor)

        self.assertIn(ec_11134, graph[fluvastatin])
        self.assertIn(ec_11188, graph[fluvastatin])
        self.assertIn(ec_11134, graph[avorastatin])
        self.assertIn(ec_11188, graph[avorastatin])
        self.assertIn(ec_11134, graph[synthetic_statin])
        self.assertIn(ec_11188, graph[synthetic_statin])
        self.assertIn(ec_11134, graph[statin])
        self.assertIn(ec_11188, graph[statin])
        self.assertIn(ec_11134, graph[mevinolinic_acid])
        self.assertIn(ec_11188, graph[mevinolinic_acid])
        self.assertIn(ec_11134, graph[hmgcr_inhibitor])
        self.assertIn(ec_11188, graph[hmgcr_inhibitor])

        self.assertEqual(9, graph.number_of_nodes())
        self.assertEqual(18, graph.number_of_edges())

        infer_child_relations(graph, ec_11134)

        self.assertIn(hmgcr, graph[fluvastatin])
        self.assertIn(hmgcr, graph[avorastatin])
        self.assertIn(hmgcr, graph[synthetic_statin])
        self.assertIn(hmgcr, graph[statin])
        self.assertIn(hmgcr, graph[mevinolinic_acid])
        self.assertIn(hmgcr, graph[hmgcr_inhibitor])

        self.assertEqual(9, graph.number_of_nodes())
        self.assertEqual(24, graph.number_of_edges())

        self.assertEqual(
            9,
            statin_graph.number_of_nodes(),
            msg="original graph nodes should not be modified",
        )
        self.assertEqual(
            8,
            statin_graph.number_of_edges(),
            msg="original graph edges should not be modified",
        )

    def test_does_not_redo(self):
        """Test that :func:`propagate_node_relations` does not add the same edges twice."""
        graph = statin_graph.copy()
        self.assertEqual(9, graph.number_of_nodes())
        self.assertEqual(8, graph.number_of_edges())

        infer_child_relations(graph, hmgcr_inhibitor)
        self.assertEqual(9, graph.number_of_nodes())
        self.assertEqual(18, graph.number_of_edges())

        infer_child_relations(graph, hmgcr_inhibitor)
        self.assertEqual(9, graph.number_of_nodes())
        self.assertEqual(18, graph.number_of_edges(), msg="edges should not be added again")


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""Convert the `CausalBionet (CBN) <http://causalbionet.com>`_ for Hipathia."""

import json
import os
import zipfile
from urllib.request import urlretrieve

import click
from pyobo.cli_utils import verbose_option
from tqdm import tqdm

import pybel
import pybel.grounding

HERE = os.path.dirname(__file__)
OUTPUT = os.path.join(HERE, 'output')
SOURCE = os.path.join(HERE, 'source')
os.makedirs(OUTPUT, exist_ok=True)

# Get and unzip this in this directory
URL = 'http://causalbionet.com/Content/jgf_bulk_files/Human-2.0.zip'
PATH = os.path.join(HERE, 'Human-2.0.zip')


@click.command()
@verbose_option
def main():
    """Convert all CBN graphs to Hipathia."""
    if not os.path.exists(PATH):
        urlretrieve(URL, PATH)

    if not os.path.exists(SOURCE):
        with zipfile.ZipFile(PATH) as file:
            file.extractall(SOURCE)

    for filename in tqdm(os.listdir(SOURCE)):
        path = os.path.join(SOURCE, filename)
        with open(path) as f:
            cbn_jgif_dict = json.load(f)

        graph = pybel.from_cbn_jgif(cbn_jgif_dict)
        graph = pybel.grounding.ground(graph)
        pybel.to_hipathia(graph, OUTPUT)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""HGVS Variants.

For example, the BEL term :code:`p(HGNC:GSK3B, var(p.Gly123Arg))` is translated to the following internal DSL:

.. code-block:: python

   from pybel.dsl import Protein, Hgvs
   gsk3b_variant = Protien(namespace='HGNC', name='GSK3B', variants=Hgvs('p.Gly123Arg'))

Further, the shorthand for protein substitutions, :class:`pybel.dsl.ProteinSubstitution`, can be used to produce the
same result, as it inherits from :class:`pybel.dsl.Hgvs`:

.. code-block:: python

   from pybel.dsl import Protein, ProteinSubstitution
   gsk3b_variant = Protien(namespace='HGNC', name='GSK3B', variants=ProteinSubstitution('Gly', 123, 'Arg'))

Either way, the resulting object can be used like a dict that looks like:

.. code-block:: python

    from pybel.constants import *

    {
        FUNCTION: PROTEIN,
        NAMESPACE: 'HGNC',
        NAME: 'GSK3B',
        VARIANTS: [
            {
                KIND: HGVS,
                IDENTIFIER: 'p.Gly123Arg',
            },
        ],
    }

.. seealso::

    - BEL 2.0 specification on `variants <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_variant_var>`_
    - HGVS `conventions <http://www.hgvs.org/mutnomen/recs.html>`_
    - PyBEL module :py:class:`pybel.parser.modifiers.get_hgvs_language`
"""

from pyparsing import ParserElement, Word, alphanums

from ..utils import nest, one_of_tags, quote
from ...constants import HGVS, KIND

__all__ = [
    "get_hgvs_language",
]

variant_tags = one_of_tags(tags=["var", "variant"], canonical_tag=HGVS, name=KIND)
variant_characters = Word(alphanums + "._*=?>")


def get_hgvs_language() -> ParserElement:
    """Build a HGVS :class:`pyparsing.ParseElement`."""
    hgvs = (variant_characters | quote)(HGVS)
    language = variant_tags + nest(hgvs)
    return language



# -*- coding: utf-8 -*-

"""Pre-defined predicates for nodes."""

from functools import wraps
from typing import Callable, Iterable

from ..typing import NodePredicate, NodePredicates
from ...graph import BELGraph
from ....dsl import BaseEntity

__all__ = [
    "node_predicate",
    "invert_node_predicate",
    "concatenate_node_predicates",
    "true_node_predicate",
    "false_node_predicate",
]


def node_predicate(f: Callable[[BaseEntity], bool]) -> NodePredicate:  # noqa: D202
    """Tag a node predicate that takes a dictionary to also accept a pair of (BELGraph, node).

    Apply this as a decorator to a function that takes a single argument, a PyBEL node, to make
    sure that it can also accept a pair of arguments, a BELGraph and a PyBEL node as well.
    """

    @wraps(f)
    def wrapped(*args):
        x = args[0]

        if isinstance(x, BELGraph):
            return f(args[1], *args[2:])
        elif isinstance(x, BaseEntity):
            return f(*args)
        else:
            raise TypeError

    return wrapped


def invert_node_predicate(f: NodePredicate) -> NodePredicate:  # noqa: D202
    """Build a node predicate that is the inverse of the given node predicate."""

    def inverse_predicate(graph: BELGraph, node: BaseEntity) -> bool:
        """Return the inverse of the enclosed node predicate applied to the graph and node."""
        return not f(graph, node)

    return inverse_predicate


def concatenate_node_predicates(node_predicates: NodePredicates) -> NodePredicate:
    """Concatenate multiple node predicates to a new predicate that requires all predicates to be met.

    Example usage:

    >>> from pybel import BELGraph
    >>> from pybel.dsl import Protein
    >>> from pybel.struct.filters import not_gene, not_rna
    >>> app_protein = Protein(name='APP', namespace='hgnc', identifier='620')
    >>> app_rna = app_protein.get_rna()
    >>> app_gene = app_rna.get_gene()
    >>> graph = BELGraph()
    >>> _ = graph.add_transcription(app_gene, app_rna)
    >>> _ = graph.add_translation(app_rna, app_protein)
    >>> node_predicate = concatenate_node_predicates([not_rna, not_gene])
    >>> assert node_predicate(graph, app_protein)
    >>> assert not node_predicate(graph, app_rna)
    >>> assert not node_predicate(graph, app_gene)
    """
    # If a predicate outside a list is given, just return it
    if not isinstance(node_predicates, Iterable):
        return node_predicates

    node_predicates = tuple(node_predicates)

    # If only one predicate is given, don't bother wrapping it
    if 1 == len(node_predicates):
        return node_predicates[0]

    def concatenated_node_predicate(graph: BELGraph, node: BaseEntity) -> bool:
        """Pass only for a nodes that pass all enclosed predicates."""
        return all(node_predicate(graph, node) for node_predicate in node_predicates)

    return concatenated_node_predicate


@node_predicate
def true_node_predicate(_: BaseEntity) -> bool:
    """Return true for all nodes.

    Given BEL graph :code:`graph`, applying :func:`true_predicate` with a predicate on the nodes iterable
    as in :code:`filter(keep_node_permissive, graph)` will result in the same iterable as iterating directly over a
    :class:`BELGraph`
    """
    return True


@node_predicate
def false_node_predicate(_: BaseEntity) -> bool:
    """Return false for all nodes."""
    return False



# -*- coding: utf-8 -*-

"""Types for filters."""

from typing import Callable, Iterable, Tuple, Union

from ..graph import BELGraph
from ...dsl import BaseEntity

__all__ = [
    "NodePredicate",
    "NodePredicates",
    "EdgeTuple",
    "EdgeIterator",
    "EdgePredicate",
    "EdgePredicates",
]

NodePredicate = Callable[[BELGraph, BaseEntity], bool]
NodePredicates = Union[NodePredicate, Iterable[NodePredicate]]

EdgeTuple = Tuple[BaseEntity, BaseEntity, str]
EdgeIterator = Iterable[EdgeTuple]

EdgePredicate = Callable[[BELGraph, BaseEntity, BaseEntity, str], bool]
EdgePredicates = Union[EdgePredicate, Iterable[EdgePredicate]]



# -*- coding: utf-8 -*-

"""Example BEL graphs for testing PyBEL-CX."""

from pybel import BELGraph
from pybel.constants import CITATION_TYPE_OTHER, HAS_VARIANT
from pybel.dsl import (
    ComplexAbundance,
    Protein,
    abundance,
    activity,
    bioprocess,
    gene,
    gmod,
    named_complex_abundance,
    pathology,
    pmod,
    protein_fusion,
    protein_substitution,
    reaction,
    rna,
)
from pybel.language import citation_dict

example_graph = BELGraph()
example_graph.annotation_list["Confidence"] = {"Low", "Medium", "High", "Very High"}
example_graph.annotation_pattern["Number"] = r"\d+"
example_graph.annotation_pattern["Species"] = r"\d+"

ptk2 = Protein(namespace="hgnc", name="PTK2", variants=pmod("Ph", "Tyr", 925))
mapk1 = Protein(namespace="hgnc", name="MAPK1")
mapk3 = Protein(namespace="hgnc", name="MAPK3")
grb2 = Protein(namespace="hgnc", name="GRB2")
sos1 = Protein(namespace="hgnc", name="SOS1")
ptk2_rgb2_sos1 = ComplexAbundance([mapk1, grb2, sos1])

ras_family = Protein(namespace="fplx", name="RAS")
pi3k_complex = named_complex_abundance(namespace="fplx", name="p85/p110 PI3Kinase Complex")

kinase_activity = activity("kin")
catalytic_activity = activity("cat")
gtp_activity = activity("gtp")

c1 = "10446041"
e1 = (
    "FAK also combines with, and may activate, phosphoinositide 3-OH kinase (PI 3-kinase), either directly or "
    "through the Src kinase (13). Finally, there is evidence that Src phosphorylates FAK at Tyr925, creating a"
    " binding site for the complex of the adapter Grb2 and Ras guanosine 5'-triphosphate exchange factor mSOS (10)."
    " These interactions link FAK to signaling pathways that modify the cytoskeleton and activate mitogen-activated"
    " protein kinase (MAPK) cascades (Fig. 3A)."
)

e1 = str(hash(e1))

"""
p(HGNC:PTK2,pmod(P,Y,925)) increases kin(p(HGNC:MAPK1))
complex(p(HGNC:PTK2),p(HGNC:GRB2),p(HGNC:SOS1)) increases gtp(p(SFAM:"RAS Family"))
gtp(p(SFAM:"RAS Family")) increases kin(p(HGNC:MAPK1))
kin(p(HGNC:PTK2)) increases complex(p(HGNC:PTK2),p(HGNC:GRB2),p(HGNC:SOS1))
kin(p(HGNC:PTK2)) increases kin(p(HGNC:MAPK1))
kin(p(HGNC:PTK2)) increases kin(p(HGNC:MAPK3))
kin(p(HGNC:PTK2)) increases kin(complex(SCOMP:"p85/p110 PI3Kinase Complex"))
"""

example_graph.add_increases(ptk2, mapk1, evidence=e1, citation=c1, target_modifier=kinase_activity)
example_graph.add_increases(ptk2_rgb2_sos1, ras_family, evidence=e1, citation=c1, target_modifier=gtp_activity)
example_graph.add_increases(
    ras_family,
    mapk1,
    evidence=e1,
    citation=c1,
    source_modifier=gtp_activity,
    target_modifier=kinase_activity,
)
example_graph.add_increases(ptk2, ptk2_rgb2_sos1, evidence=e1, citation=c1, source_modifier=kinase_activity)
example_graph.add_increases(
    ptk2,
    mapk1,
    evidence=e1,
    citation=c1,
    source_modifier=kinase_activity,
    target_modifier=kinase_activity,
)
example_graph.add_increases(
    ptk2,
    mapk3,
    evidence=e1,
    citation=c1,
    source_modifier=kinase_activity,
    target_modifier=kinase_activity,
)
example_graph.add_increases(
    ptk2,
    pi3k_complex,
    evidence=e1,
    citation=c1,
    source_modifier=kinase_activity,
    target_modifier=kinase_activity,
)

"""
SET Evidence=" two common MTHFR polymorphisms, namely 677C>T (Ala222Val) and 1298A>C (Glu429Ala), are known to reduce MTHFR activity. \
It has been shown that the MTHFR 677T allele is associated with increased total plasma Hcy levels (tHcy) and decreased serum folate levels, mainly in 677TT homozygous subjects.\
the MTHFR 677C>T polymorphism as a candidate AD risk factor"

SET Subgraph = "Epigenetic modification subgraph"
g(HGNC:MTHFR,sub(C,677,T)) =| p(HGNC:MTHFR)
g(HGNC:MTHFR,sub(A,1298,C)) =| p(HGNC:MTHFR)
g(HGNC:MTHFR,sub(C,677,T)) neg a(CHEBI:"folic acid")
g(HGNC:MTHFR,sub(C,677,T)) pos path(MESH:"Alzheimer Disease")
"""

c2 = "21119889"
e2 = "Two common MTHFR polymorphisms, namely 677C>T (Ala222Val) and 1298A>C (Glu429Ala), are known to reduce MTHFR activity. \
It has been shown that the MTHFR 677T allele is associated with increased total plasma Hcy levels (tHcy) and decreased serum folate levels, mainly in 677TT homozygous subjects.\
the MTHFR 677C>T polymorphism as a candidate AD risk factor"
e2 = str(hash(e2))

mthfr = Protein(namespace="hgnc", name="MTHFR")
mthfr_c677t = Protein(namespace="hgnc", name="MTHFR", variants=[protein_substitution("Ala", 222, "Val")])
mthfr_a1298c = Protein(namespace="hgnc", name="MTHFR", variants=[protein_substitution("Glu", 429, "Ala")])
folic_acid = abundance("CHEBI", "folic acid")
alzheimer_disease = pathology("MESH", "Alzheimer Disease")

example_graph.add_decreases(mthfr_c677t, mthfr, citation=c2, evidence=e2, target_modifier=activity())
example_graph.add_decreases(mthfr_a1298c, mthfr, citation=c2, evidence=e2, target_modifier=activity())
example_graph.add_negative_correlation(mthfr_c677t, folic_acid, citation=c2, evidence=e2)
example_graph.add_positive_correlation(mthfr_c677t, alzheimer_disease, citation=c2, evidence=e2)

c3 = "17948130"
e3 = (
    "A polymorphism in the NDUFB6 promoter region that creates a possible DNA methylation site (rs629566, A/G) was "
    "associated with a decline in muscle NDUFB6 expression with age. Although young subjects with the rs629566 G/G "
    "genotype exhibited higher muscle NDUFB6 expression, this genotype was associated with reduced expression in"
    " elderly subjects. This was subsequently explained by the finding of increased DNA methylation in the promoter "
    "of elderly, but not young, subjects carrying the rs629566 G/G genotype. Furthermore, the degree of DNA"
    " methylation correlated negatively with muscle NDUFB6 expression, which in turn was associated with insulin "
    "sensitivity."
)
e3 = str(hash(e3))

rs629566 = gene("DBSNP", "rs629566", variants=[gmod("Me")])
ndufb6_gene = gene("HGNC", "NDUFB6")
ndufb6_rna = rna("HGNC", "NDUFB6")

example_graph.add_unqualified_edge(ndufb6_gene, rs629566, HAS_VARIANT)
example_graph.add_negative_correlation(
    rs629566,
    ndufb6_rna,
    citation=c3,
    evidence=e3,
    annotations={"Confidence": "Low", "Number": "50"},
)

"""
SET Evidence = "% Entrez Gene summary: Rat: SUMMARY: precursor protein of kinin which is found in plasma; cysteine protease inhibitor and a major acute phase reactant [RGD] OMIM summary: (summary is not available from this source) kininogens; Endogenous peptides present in most body fluids. Certain enzymes convert them to active kinins which are involved in inflammation, blood clotting, complement reactions, etc. Kininogens belong to the cystatin superfamily. They are cysteine proteinase inhibitors. High-molecular-weight kininogen (hmwk) is split by plasma kallikrein to produce bradykinin. Low-molecular-weight kininogen (lmwk) is split by tissue kallikrein to produce kallidin. kinins; Inflammatory mediators that cause dilation of blood vessels and altered vascular permeability.  Kinins are small peptides produced from kininogen by kallikrein and are broken down by kininases. Act on phospholipase and increase arachidonic acid release and thus prostaglandin (PGE2) production. bradykinin; Vasoactive nonapeptide (RPPGFSPFR) formed by action of proteases on kininogens. Very similar to kallidin (which has the same sequence but with an additional N terminal lysine). Bradykinin is a very potent vasodilator and increases permeability of post capillary venules, it acts on endothelial cells to activate phospholipase A2. It is also spasmogenic for some smooth muscle and will cause pain. kallidin; Decapeptide (lysyl bradykinin, amino acid sequence KRPPGFSPFR) produced in kidney. Like bradykinin, an inflammatory mediator (a kinin), causes dilation of renal blood vessels and increased water excretion."
SET Species = 9606
SET Citation = {"Other","Genstruct Kininogen Overview","Genstruct Kininogen Overview","","",""}

bp(GOBP:"inflammatory response") increases rxn(reactants(p(HGNC:KNG1)),products(a(SCHEM:Kallidin)))
path(SDIS:"tissue damage") increases rxn(reactants(p(HGNC:KNG1)),products(a(SCHEM:Kallidin)))
a(SCHEM:Kallidin) increases cat(p(HGNC:BDKRB1))
cat(p(HGNC:BDKRB1)) increases cat(p(SFAM:"PLA2 Family"))
"""

c4 = citation_dict(namespace=CITATION_TYPE_OTHER, identifier="Genstruct Reference")
e4 = "% Entrez Gene summary: Rat: SUMMARY: precursor protein of kinin which is found in plasma; cysteine protease inhibitor and a major acute phase reactant [RGD] OMIM summary: (summary is not available from this source) kininogens; Endogenous peptides present in most body fluids. Certain enzymes convert them to active kinins which are involved in inflammation, blood clotting, complement reactions, etc. Kininogens belong to the cystatin superfamily. They are cysteine proteinase inhibitors. High-molecular-weight kininogen (hmwk) is split by plasma kallikrein to produce bradykinin. Low-molecular-weight kininogen (lmwk) is split by tissue kallikrein to produce kallidin. kinins; Inflammatory mediators that cause dilation of blood vessels and altered vascular permeability.  Kinins are small peptides produced from kininogen by kallikrein and are broken down by kininases. Act on phospholipase and increase arachidonic acid release and thus prostaglandin (PGE2) production. bradykinin; Vasoactive nonapeptide (RPPGFSPFR) formed by action of proteases on kininogens. Very similar to kallidin (which has the same sequence but with an additional N terminal lysine). Bradykinin is a very potent vasodilator and increases permeability of post capillary venules, it acts on endothelial cells to activate phospholipase A2. It is also spasmogenic for some smooth muscle and will cause pain. kallidin; Decapeptide (lysyl bradykinin, amino acid sequence KRPPGFSPFR) produced in kidney. Like bradykinin, an inflammatory mediator (a kinin), causes dilation of renal blood vessels and increased water excretion."
e4 = str(hash(e4))

bdkrb1 = Protein(namespace="hgnc", name="BDKRB1")
inflammatory_process = bioprocess("GO", "inflammatory process")
kng1 = Protein(namespace="hgnc", name="KNG1")
kallidin = abundance("CHEBI", "Kallidin")
pla2_family = Protein("SFAM", "PLA2 Family")
kng1_to_kallidin = reaction(reactants=[kng1], products=[kallidin])

example_graph.add_increases(inflammatory_process, kng1_to_kallidin, citation=c4, evidence=e4)
example_graph.add_increases(kallidin, bdkrb1, citation=c4, evidence=e4, target_modifier=catalytic_activity)
example_graph.add_increases(
    bdkrb1,
    pla2_family,
    citation=c4,
    evidence=e4,
    source_modifier=catalytic_activity,
    target_modifier=catalytic_activity,
)

c5 = "10866298"
e5 = "We found that PD180970 inhibited in vivo tyrosine phosphorylation of p210Bcr-Abl (IC50 = 170 nM) and the p210BcrAbl substrates Gab2 and CrkL (IC50 = 80 nM) in human K562 chronic myelogenous leukemic cells. In vitro, PD180970 potently inhibited autophosphorylation of p210Bcr-Abl (IC50 = 5 nM) and the kinase activity of purified recombinant Abl tyrosine kinase (IC50 = 2.2 nM)."

"""
SET Species = 9606
SET Citation = {"PubMed","Cancer Res 2000 Jun 15 60(12) 3127-31","10866298","","",""}

kin(p(HGNC:BCR,fus(HGNC:ABL1))) directlyIncreases p(HGNC:CRKL,pmod(P,Y))
kin(p(HGNC:BCR,fus(HGNC:ABL1))) directlyIncreases p(HGNC:GAB2,pmod(P,Y))
"""

bcr_abl1_fus = protein_fusion(
    partner_5p=Protein(namespace="hgnc", name="BCR"),
    partner_3p=Protein(namespace="hgnc", name="ABL1"),
)
crkl_ph = Protein(namespace="hgnc", name="CRKL", variants=[pmod("Ph", "Tyr")])
gab2_ph = Protein(namespace="hgnc", name="GAB2", variants=[pmod("Ph", "Tyr")])

example_graph.add_directly_increases(
    bcr_abl1_fus,
    crkl_ph,
    citation=c5,
    evidence=e5,
    annotations={"Species": "9606", "Confidence": "High"},
    source_modifier=kinase_activity,
)
example_graph.add_directly_increases(
    bcr_abl1_fus,
    gab2_ph,
    citation=c5,
    evidence=e5,
    annotations={"Species": "9606"},
    source_modifier=kinase_activity,
)



# -*- coding: utf-8 -*-

"""Importer for Hetionet JSON."""

import unittest

import pybel.io.hetionet.constants as hioc
from pybel import dsl
from pybel.io.hetionet import from_hetionet_json


class TestHetionet(unittest.TestCase):
    def test_import(self):
        self._help(
            h_type=hioc.ANATOMY,
            h_dsl=dsl.Population,
            h_namespace="uberon",
            h_id="UBERON:1",
            h_name="anatomy1",
            kind="upregulates",
            t_type=hioc.GENE,
            t_dsl=dsl.Rna,
            t_namespace="ncbigene",
            t_id="1",
            t_name="gene1",
        )

    def _help(
        self,
        h_type,
        h_dsl,
        h_namespace,
        h_id,
        h_name,
        kind,
        t_type,
        t_dsl,
        t_namespace,
        t_id,
        t_name,
    ):
        source = dict(kind=h_type, identifier=h_id, name=h_name)
        target = dict(kind=t_type, identifier=t_id, name=t_name)
        edge = dict(
            source_id=(h_type, h_id),
            kind=kind,
            target_id=(t_type, t_id),
            data={},
        )

        if h_id.lower().startswith("{}:".format(h_namespace.lower())):
            h_id = h_id[len(h_namespace) + 1 :]

        if t_id.lower().startswith("{}:".format(t_namespace.lower())):
            t_id = t_id[len(t_namespace) + 1 :]

        graph = from_hetionet_json(dict(nodes=[source, target], edges=[edge]), use_tqdm=False)
        source_node = h_dsl(namespace=h_namespace, identifier=h_id, name=h_name)
        self.assertIn(source_node, graph, msg="Nodes: {}".format(list(graph)))
        target_node = t_dsl(namespace=t_namespace, identifier=t_id, name=t_name)
        self.assertIn(target_node, graph, msg="Nodes: {}".format(list(graph)))
        self.assertEqual(2, graph.number_of_nodes())

        self.assertIn(target_node, graph[source_node])
